/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include <mpi.h>
#include "graleconfig.h"
#include "renderserver.h"
#include "gravitationallens.h"
#include <serut/memoryserializer.h>
#include <enut/tcpv4socket.h>
#include <enut/tcppacketsocket.h>
#include <enut/packet.h>
#include <enut/socketwaiter.h>
#include <stdlib.h>
#include <signal.h>
#include <iostream>
#include <string>

#include "debugnew.h"

using namespace grale;

class CurrentLensInfo;

class CurrentLensInfo
{
public:
	CurrentLensInfo(int numx, int numy)									
	{ 									
		xsteps = numx; 
		ysteps = numy; 
		
		linemarkers = new bool [ysteps]; 
		
		for (int i = 0 ; i < ysteps ; i++) 
			linemarkers[i] = false; 
		calculatedlines = 0; 
		requestedlines = 0; 
		imagesize = 5*sizeof(double)*numx*numy+sizeof(int32_t); 
		
		image = new uint8_t [imagesize]; 
	
		serut::MemorySerializer mser(0, 0, image, sizeof(int32_t)); 
		mser.writeInt32(RENDERSERVER_COMMAND_IMAGEPLANE); 
	}

	~CurrentLensInfo()							
	{ 
		delete [] linemarkers; 
		delete [] image; 
	}
	
	int getXSteps() const							{ return xsteps; }
	int getTotalLines() const						{ return ysteps; }
	int getCalculatedLines() const						{ return calculatedlines; }

	void reenterLine(int linenr)						{ requestedlines--; linemarkers[linenr] = false; }
	int requestLine()							{ if (requestedlines == ysteps) return -1; for (int i = 0 ; i < ysteps ; i++) if (!linemarkers[i]) { linemarkers[i] = true; requestedlines++; return i; }  return -1; }
	
	void storeCalculatedLine(int l,const uint8_t *buf)			{ memcpy(image+sizeof(int32_t)+l*5*sizeof(double)*xsteps,buf,5*sizeof(double)*xsteps); calculatedlines++; }
	uint8_t *getImagePlane()						{ return image; }
	int getImagePlaneSize() const						{ return imagesize; }
private:
	int xsteps, ysteps;

	bool *linemarkers;
	int calculatedlines;
	int requestedlines;

	uint8_t *image;
	int imagesize;
};

#define TIMEOUT 300

bool readNextPacket(nut::TCPPacketSocket *pConn, nut::Packet &pack)
{
	time_t startTime = time(NULL);
	bool done = false;

	while (!done)
	{
		time_t curTime = time(NULL);

		if (curTime - startTime > TIMEOUT)
		{
			std::cerr << "Timeout while waiting for incoming packet from client" << std::endl;
			return false;
		}
		if (!pConn->waitForData(10, 0))
		{
			std::cerr << "Error while waiting for incoming packet from client: " << pConn->getErrorString() << std::endl;
			return false;
		}
		if (!pConn->poll())
		{
			std::cerr << "Error while polling for incoming data on client connection: " << pConn->getErrorString() << std::endl;
			return false;
		}
		
		if (pConn->read(pack))
			done = true;
	}
	
	return true;
}

int serverLoop(int portNumber, int numProcs)
{
	nut::TCPPacketSocket *pConn;

	// Listen for incoming connections
	{
			
		nut::TCPv4Socket listenSocket;

		if (!listenSocket.create(portNumber))
		{
			std::cerr << "Error creating server at port " << portNumber << ":" << std::endl;
			std::cerr << listenSocket.getErrorString() << std::endl;
			MPI_Abort(MPI_COMM_WORLD, -1);
			return -1;
		}

		if (!listenSocket.listen(0))
		{
			std::cerr << "Couldn't set socket in listen mode:" << std::endl;
			std::cerr << listenSocket.getErrorString() << std::endl;
			MPI_Abort(MPI_COMM_WORLD, -1);
			return -1;
		}

		std::cout << "Listen socket created at port " << listenSocket.getLocalPortNumber() << std::endl;

		nut::SocketWaiter waiter;

		waiter.addSocket(listenSocket);

		if (!waiter.wait(-1, -1))
		{
			std::cerr << "Error waiting for incoming connection" << std::endl;
			MPI_Abort(MPI_COMM_WORLD, -1);
			return -1;
		}

		// Incoming connection: accept it

		nut::TCPSocket *pSock;

		if (listenSocket.accept(&pSock))
			pConn = new nut::TCPPacketSocket(pSock, true, false, RENDERSERVER_MAXPACKSIZE, sizeof(uint32_t), RENDERSERVER_PACKID);
		else
		{
			std::cerr << "Error accepting incoming connection" << std::endl;
			MPI_Abort(MPI_COMM_WORLD, -1);
			return -1;
		}
	}

	// Got connection, read client identifier
	
	{
		nut::Packet pack;

		if (!readNextPacket(pConn, pack))
		{
			MPI_Abort(MPI_COMM_WORLD, -1);
			return -1;
		}
		if (pack.getLength() != sizeof(uint32_t))
		{
			std::cerr << "Read packet of invalid length" << std::endl;
			MPI_Abort(MPI_COMM_WORLD, -1);
			return -1;
		}
			
		serut::MemorySerializer mser(pack.getData(), sizeof(int32_t), 0, 0);

		int32_t id;

		mser.readInt32(&id);
		if (id != RENDERSERVER_COMMAND_CLIENT)
		{
			std::cerr << "Incoming connection is not a rendering client" << std::endl;
			MPI_Abort(MPI_COMM_WORLD, -1);
			return -1;
		}	
	}

	// Send server accept reply
	
	{
		uint8_t buf[sizeof(uint32_t)];
		serut::MemorySerializer mser(0, 0, buf, sizeof(int32_t));

		mser.writeInt32(RENDERSERVER_COMMAND_ACCEPT);

		pConn->write(buf, sizeof(uint32_t));
	}

	nut::Packet lensPacket;
	CurrentLensInfo *pLensInfo = 0;
	int32_t width, height;

	std::cout << "Reading lens info" << std::endl;

	// read the packet containing the info about what should be rendered
	
	{
		if (!readNextPacket(pConn, lensPacket))
		{
			MPI_Abort(MPI_COMM_WORLD, -1);
			return -1;
		}
		if (lensPacket.getLength() <= sizeof(uint32_t)*3+6*sizeof(double))
		{
			std::cerr << "Lens packet has invalid length" << std::endl;
			MPI_Abort(MPI_COMM_WORLD, -1);
			return -1;
		}

		serut::MemorySerializer mser(lensPacket.getData(), lensPacket.getLength(), 0, 0);

		int32_t id;

		mser.readInt32(&id);

		if (id != RENDERSERVER_COMMAND_LENS)
		{
			std::cerr << "Lens packet contains invalid identifier" << std::endl;
			MPI_Abort(MPI_COMM_WORLD, -1);
			return -1;
		}

		mser.readInt32(&width);
		mser.readInt32(&height);

		// std::cout << width << "x" << height << std::endl;

		pLensInfo = new CurrentLensInfo(width, height);

		// TODO: check other params and/or lens itself?
	}

	// Send the lens info to the helpers

	for (int i = 1 ; i < numProcs ; i++)
	{
		int cmd = RENDERSERVER_COMMAND_LENS;
		int packetLength = lensPacket.getLength();

		MPI_Send(&cmd, 1, MPI_INT, i, 0, MPI_COMM_WORLD);
		MPI_Send(&packetLength, 1, MPI_INT, i, 0, MPI_COMM_WORLD);
		MPI_Send((void *)lensPacket.getData(), packetLength, MPI_BYTE, i, 0, MPI_COMM_WORLD);
	}

	std::vector<double> resultBuffer(width*5);
	std::vector<int> renderInfo(numProcs);

	for (int i = 0 ; i < renderInfo.size() ; i++)
		renderInfo[i] = -1;

	while (pLensInfo->getTotalLines() != pLensInfo->getCalculatedLines())
	{
		bool done = false;
		for (int i = 1 ; !done && i < renderInfo.size() ; i++)
		{
			if (renderInfo[i] == -1)
			{
				int lineNumber = pLensInfo->requestLine();

				if (lineNumber == -1)
					done = true;
				else
				{
					int command = RENDERSERVER_COMMAND_RENDER;
					
					// std::cout << "Writing line number " << lineNumber << " to process " << i << std::endl;
					MPI_Send(&command, 1, MPI_INT, i, 0, MPI_COMM_WORLD);
					MPI_Send(&lineNumber, 1, MPI_INT, i, 0, MPI_COMM_WORLD);

					renderInfo[i] = lineNumber;
				}
			}
		}

		// Wait for a received line

		// std::cout << "Waiting for a received line " << std::endl;

		int flag;
		MPI_Status status;

		MPI_Iprobe(MPI_ANY_SOURCE, 0, MPI_COMM_WORLD, &flag, &status);

		if (flag)
		{
			MPI_Recv(&(resultBuffer[0]), resultBuffer.size(), MPI_DOUBLE, MPI_ANY_SOURCE, 0, MPI_COMM_WORLD, &status);

			// std::cout << "Source is: " << status.MPI_SOURCE << std::endl;

			int line = renderInfo[status.MPI_SOURCE];

			pLensInfo->storeCalculatedLine(line, (const uint8_t *)&(resultBuffer[0]));
			renderInfo[status.MPI_SOURCE] = -1;

			// Send percentage to client

			int pct = (int)(100.0*((double)pLensInfo->getCalculatedLines()/(double)pLensInfo->getTotalLines())+0.5);

			if (pct > 100)
				pct = 100;

			uint8_t data[sizeof(int32_t)*2];
			serut::MemorySerializer mser(0,0,data,sizeof(int32_t)*2);

			mser.writeInt32(RENDERSERVER_COMMAND_PCT);
			mser.writeInt32(pct);
			
			if (!pConn->write(data, sizeof(int32_t)*2))
			{
				std::cerr << "Error writing progress to client: " << pConn->getErrorString() << std::endl;
				MPI_Abort(MPI_COMM_WORLD, -1);
				return -1;
			}
		}
		else
		{
			// Wait a bit, and read any incoming echo packets from the client

			nut::SocketWaiter waiter;

			waiter.addSocket(*pConn);
			waiter.wait(0, 100000);

			if (pConn->isDataAvailable())
			{
				nut::Packet dummy;

				pConn->poll();

				// Just ignore incoming data
				//std::cout << "Reading and ignoring incoming packet" << std::endl;
				pConn->read(dummy);
			}

			if (time(0) - pConn->getLastWriteTime() > 10)
			{
				// std::cout << "Sending keepalive to client" << std::endl;

				uint8_t data[sizeof(int32_t)];
				serut::MemorySerializer mser(0,0,data,sizeof(int32_t));

				mser.writeInt32(RENDERSERVER_COMMAND_ECHO);
				
				if (!pConn->write(data, sizeof(int32_t)))
				{
					std::cerr << "Error writing keepalive to client: " << pConn->getErrorString() << std::endl;
					MPI_Abort(MPI_COMM_WORLD, -1);
					return -1;
				}
			}
			if (time(0) - pConn->getLastReadTime() > 1200)
			{
				std::cerr << "Timeout on client connection" << std::endl;
				MPI_Abort(MPI_COMM_WORLD, -1);
				return -1;
			}
		}
	}

	// Rendered everyting, send data to client

	// First, write "100%"
	{
		uint8_t data[sizeof(int32_t)*2];
		serut::MemorySerializer mser(0,0,data,sizeof(int32_t)*2);

		mser.writeInt32(RENDERSERVER_COMMAND_PCT);
		mser.writeInt32(100); 
		
		if (!pConn->write(data, sizeof(int32_t)*2))
		{
			std::cerr << "Error writing progress to client: " << pConn->getErrorString() << std::endl;
			MPI_Abort(MPI_COMM_WORLD, -1);
			return -1;
		}
	}

	// Then, write result
	{
		if (!pConn->write(pLensInfo->getImagePlane(), pLensInfo->getImagePlaneSize()))
		{
			std::cerr << "Error writing image plane to client: " << pConn->getErrorString() << std::endl;
			MPI_Abort(MPI_COMM_WORLD, -1);
			return -1;
		}


		// wait for acknowledgement

		bool done = false;
		while (!done)
		{
			nut::Packet ackPacket;

			if (!readNextPacket(pConn, ackPacket)) 
				done = true; // ignore the error and quit
			else
			{
				if (ackPacket.getLength() == sizeof(int32_t))
				{
					serut::MemorySerializer mser(ackPacket.getData(), sizeof(int32_t), 0, 0);
					int32_t cmd;

					mser.readInt32(&cmd);
					if (cmd == RENDERSERVER_COMMAND_ACKIMAGEPLANE)
					{
						std::cout << "Got image plane acknowledgement" << std::endl;
						done = true;
					}
				}
			}
		}
	}
	std::cout << "Exiting" << std::endl;

	return 0;
}

int clientLoop()
{
	// std::cout << "Client started" << std::endl;

	int cmd = 0;
	int length;
	MPI_Status status;

	MPI_Recv(&cmd, 1, MPI_INT, 0, 0, MPI_COMM_WORLD, &status);

	//std::cout << "Read RENDERSERVER_COMMAND_LENS?" << std::endl;
	//std::cout << "cmd = " << cmd << std::endl;

	if (cmd != RENDERSERVER_COMMAND_LENS)
	{
		std::cerr << "Didn't receive render command, exiting" << std::endl;
		return -1;
	}

	MPI_Recv(&length, 1, MPI_INT, 0, 0, MPI_COMM_WORLD, &status);
	
	//std::cout << "Expecting a packet of size " << length << std::endl;
	
	std::vector<uint8_t> buffer(length);
	serut::MemorySerializer mser(&(buffer[0]), length, 0, 0);

	MPI_Recv(&(buffer[0]), length, MPI_BYTE, 0, 0, MPI_COMM_WORLD, &status);
	// TODO: error check?

	int32_t cmdDummy;
	int32_t numx, numy;
	Vector2D<double> topright, bottomleft;
	double D_s, D_ds;

	mser.readInt32(&cmdDummy); // contains RENDERSERVER_COMMAND_LENS again

	//std::cout << "cmdDummy is " << cmdDummy << std::endl;

	mser.readInt32(&numx);
	mser.readInt32(&numy);
	mser.readDoubles(bottomleft.getComponents(), 2);
	mser.readDoubles(topright.getComponents(), 2);
	mser.readDouble(&D_s);
	mser.readDouble(&D_ds);

	GravitationalLens *pLens = 0;
	std::string errstr;

	GravitationalLens::read(mser, &pLens, errstr);

	//std::cout << "Width = " << numx << " Height = " << numy << std::endl;
	//std::cout << "Lens: " << pLens << " Error: " << errstr << std::endl; 

	// Reserve some space for the render data
	
	std::vector<double> line(numx*5);
	double dx = (topright.getX()-bottomleft.getX())/((double)(numx-1));
	double dy = (topright.getY()-bottomleft.getY())/((double)(numy-1));

	// Got the lens and the data, wait for command and execute it
	
	while (1)
	{
		MPI_Recv(&cmd, 1, MPI_INT, 0, 0, MPI_COMM_WORLD, &status);
	
		// std::cout << "Received command " << cmd << std::endl;

		if (cmd != RENDERSERVER_COMMAND_RENDER)
		{
			std::cerr << "Didn't receive render command, exiting" << std::endl;
			return 0;
		}

		int lineNumber;

		MPI_Recv(&lineNumber, 1, MPI_INT, 0, 0, MPI_COMM_WORLD, &status);

		//std::cout << "Rendering line " << lineNumber << std::endl;

		double x = bottomleft.getX();
		double y = bottomleft.getY()+(((double)lineNumber)*dy);

		for (int i = 0 ; i < numx ; i++, x+= dx)
		{
			Vector2D<double> theta(x,y);
			Vector2D<double> alpha;

			pLens->getAlphaVector(theta, &alpha);
			
			alpha *= (D_ds/D_s);

			line[5*i+0] = alpha.getX();
			line[5*i+1] = alpha.getY();
			
			double axx,ayy,axy;
			pLens->getAlphaVectorDerivatives(theta,axx,ayy,axy);
			
			line[5*i+2] = axx;
			line[5*i+3] = ayy;
			line[5*i+4] = axy;
		}

		//std::cout << "Sending result to main process" << std::endl;
		MPI_Send(&(line[0]), numx*5, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD);
		//std::cout << "Results sent successfully" << std::endl;
	}

	return 0;
}

int main2(int argc,char *argv[])
{
	if (argc != 2)
	{
		std::cerr << "Usage: " << std::endl;
		std::cerr << std::string(argv[0]) << " portnumber" << std::endl; 
		return -1;
	}

	int dummyArgc = 1;
	char dummyName[] = "dummy";
	char *dummyArg0[] = { dummyName, NULL };
	char **dummyArgv = dummyArg0;
	int nprocs, rank;

	MPI_Init(&dummyArgc, &dummyArgv);
	MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);

	if (nprocs < 2)
	{
		std::cerr << "At least two processes are required" << std::endl;
		MPI_Finalize();
		return -1;
	}

	int status;

	if (rank == 0)
	{
		status = serverLoop(atoi(argv[1]), nprocs);
	
		std::vector<MPI_Request> requests(nprocs-1);
		std::vector<MPI_Status> statusArray(nprocs-1);
		int cmd = RENDERSERVER_COMMAND_UNAVAILABLE; // just write something

		for (int i = 1 ; i < nprocs ; i++)
			MPI_Isend(&cmd, 1, MPI_INT, i, 0, MPI_COMM_WORLD, &(requests[i-1]));

		MPI_Waitall(nprocs-1, &(requests[0]), &(statusArray[0]));
	}
	else
	{
		status = clientLoop();

		char str[1024];

		gethostname(str, 1024);
		std::cout << str << std::endl;
	}
	MPI_Finalize();

	return status;
}

int main(int argc, char *argv[])
{
#ifdef WIN32
	WSADATA dat;
	WSAStartup(MAKEWORD(2,2),&dat);
#endif // WIN32
	int status = main2(argc, argv);
#ifdef WIN32
	WSACleanup();
#endif
	return status;
}