/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "gridlensinversiongenomenew.h"
#include "gridlensinversiongafactorynew.h"
#include "gridlensinversiongafactoryparamsnew.h"
#include "multipleplummerlens.h"
#include "multiplesquarelens.h"
#include "multiplegausslens.h"
#include "doublepeakdistribution.h"
#include "uniformdistribution.h"
#include "simpleuniformdistribution.h"
#include "gaussiandistribution.h"
#include "powerlawdistribution.h"
#include "jeffreysdistribution.h"
#include "peakdistribution.h"
#include "studenttdistribution.h"
#include "constants.h"
#include "lensfitnessobject.h"
#include "deflectionmatrix.h"
#include "backprojectmatrixnew.h"
#include <iostream>

#include "debugnew.h"

using namespace std;

// #define EXPERIMENTAL_MUTATION

namespace grale
{

GridLensInversionGenomeNew::GridLensInversionGenomeNew(GridLensInversionGAFactoryNew *pFactory, int numMasses, bool useMassSheet)
{
	m_pFactory = pFactory;
	
	SimpleUniformDistribution uniDist(m_pFactory->getRandomNumberGenerator());

	if (useMassSheet)
		m_sheetValue = -1;
	else
		m_sheetValue = uniDist.pickNumber();
	
	m_masses.resize(numMasses);
	m_scaleFactor = 1.0f;
	m_activeFitnessComponent = 0;

	initGenome();
}

GridLensInversionGenomeNew::GridLensInversionGenomeNew(GridLensInversionGAFactoryNew *pFactory, const std::vector<float> &masses, float sheetValue)
{
	m_pFactory = pFactory;
	m_activeFitnessComponent = 0;
	m_scaleFactor = 1.0f;
	m_masses = masses;
	m_sheetValue = sheetValue;
}

GridLensInversionGenomeNew::~GridLensInversionGenomeNew()
{
}
	
bool GridLensInversionGenomeNew::calculateFitness()
{
	// TODO:
	/*
	{
	float sheetFactor = m_pFactory->getSheetScale()*m_sheetValue;
	m_scaleFactor = 1.0f;
	calculateTotalFitness(m_scaleFactor, sheetFactor, m_fitnessValues);
	return;
	}
	*/

	initializeNewCalculation();

	// TODO: negative values
	
	float sheetFactor = m_pFactory->getSheetScale()*m_sheetValue;
	float currentBestFitness = 1e20f;
	float currentBestScaleFactor = 1.0f;

	// TODO: tweak these parameters
	float startFactor = 0.2;
	float stopFactor = 5.0;
	int numIterationSteps = 20;
	int numIterations = 5;

	float startValue = startFactor;
	float stopValue = stopFactor;

	for (int i = 0 ; i < numIterations ; i++)
	{
		float stepSize = (stopValue-startValue)/((float)(numIterationSteps-1));
		float currentFactor = startValue;

		for (int j = 0 ; j < numIterationSteps ; j++, currentFactor += stepSize)
		{
			float fitness;
			
			if (!calculateMassScaleFitness(currentFactor, sheetFactor, fitness))
				return false;

			if (fitness < currentBestFitness)
			{
				currentBestFitness = fitness;
				currentBestScaleFactor = currentFactor;
			}
		}

		startValue = MAX(startValue, currentBestScaleFactor-stepSize);
		stopValue = MIN(stopValue, currentBestScaleFactor+stepSize);
	}

	m_scaleFactor = currentBestScaleFactor;

//	std::cout << "Best scale factor is " << m_scaleFactor << std::endl;
//	std::cout << "Best fitness is " << currentBestFitness << std::endl;

	// Do the final (possibly multi-component) fitness evaluation
	calculateTotalFitness(m_scaleFactor, sheetFactor, m_fitnessValues);

//	std::cout << getFitnessDescription() << std::endl << std::endl;

	return true;
}

mogal::Genome *GridLensInversionGenomeNew::reproduce(const mogal::Genome *g) const
{
	// TODO: something better?

	std::vector<float> newMasses(m_masses.size());
	SimpleUniformDistribution uniDist(m_pFactory->getRandomNumberGenerator());
	const GridLensInversionGenomeNew *pGenome = (const GridLensInversionGenomeNew *)g;
	float otherScaleFactor = pGenome->m_scaleFactor;
	const std::vector<float> &otherMasses = pGenome->m_masses;

	for (int i = 0 ; i < m_masses.size() ; i++)
	{
		 float x = (float)uniDist.pickNumber();

		 if (x < 0.5f)
			 newMasses[i] = m_scaleFactor*m_masses[i];
		 else
			 newMasses[i] = otherScaleFactor*otherMasses[i];
	}

	// TODO: mass sheet

	return new GridLensInversionGenomeNew(m_pFactory, newMasses, m_sheetValue);
}

mogal::Genome *GridLensInversionGenomeNew::clone() const
{
	std::vector<float> newMasses(m_masses.size());

	for (int i = 0 ; i < m_masses.size() ; i++)
		newMasses[i] = m_masses[i] * m_scaleFactor;

	GridLensInversionGenomeNew *pNewGenome = new GridLensInversionGenomeNew(m_pFactory, newMasses, m_sheetValue);
	pNewGenome->setFitnessValues(m_fitnessValues);

	return pNewGenome;
}

void GridLensInversionGenomeNew::mutate()
{
	float maxValue = 0;
	for (int i = 0 ; i < m_masses.size() ; i++)
		maxValue = MAX(ABS(m_masses[i]), maxValue);

	GaussianDistribution numMutDist(1.5, 1.5, m_pFactory->getRandomNumberGenerator());

	int numMutations = (int)(numMutDist.pickNumber()+0.5);
	float mutValue = 0.2;

	if (!m_pFactory->useAbsoluteMutation())
		mutValue = 0.25*m_pFactory->getMutationAmplitude();

	const std::vector<Vector2D<float> > &centers = m_pFactory->getBasisCenters();
	float minWidth = m_pFactory->getMinimalMutationWidth();
	float maxWidth = m_pFactory->getMaximalMutationWidth();
	Vector2D<float> topRight = m_pFactory->getTopRight();
	Vector2D<float> bottomLeft = m_pFactory->getBottomLeft();
	//GaussianDistribution widthDist(0, maxWidth/5.0, m_pFactory->getRandomNumberGenerator());
	PeakDistribution widthDist(minWidth, maxWidth, m_pFactory->getRandomNumberGenerator());
	UniformDistribution xDist(bottomLeft.getX(), topRight.getX(), m_pFactory->getRandomNumberGenerator());
	UniformDistribution yDist(bottomLeft.getY(), topRight.getY(), m_pFactory->getRandomNumberGenerator());
	StudentTDistribution hDist(maxValue * mutValue, m_pFactory->getRandomNumberGenerator());
	UniformDistribution h2Dist(0, maxValue*1.2, m_pFactory->getRandomNumberGenerator());
	SimpleUniformDistribution simpDist(m_pFactory->getRandomNumberGenerator());
	UniformDistribution smallDist(0, 0.001*maxValue, m_pFactory->getRandomNumberGenerator());
	UniformDistribution genomeDist(0, m_masses.size(), m_pFactory->getRandomNumberGenerator());

	// TODO: use these
	bool absoluteMutation = m_pFactory->useAbsoluteMutation();
	float mutationAmplitude = m_pFactory->getMutationAmplitude();

	for (int i = 0 ; i < numMutations ; i++)
	{
/*		float rndValue = simpDist.pickNumber();

		if (rndValue > 0.5f)
		{*/
			Vector2D<float> center((float)xDist.pickNumber(), (float)yDist.pickNumber());
			
			float width = widthDist.pickNumber()/3.0;
			float width2 = width*width;
			float h;
			
			h = hDist.pickNumber();
	
			for (int j = 0 ; j < centers.size() ; j++)
			{
				Vector2D<float> diff = centers[j]-center;
				float distSquared = diff.getLengthSquared();

				m_masses[j] += h*EXP(-distSquared/(width2*2.0));
				if (m_masses[j] < 0)
					m_masses[j] = smallDist.pickNumber(); // avoid creating a totally zero mass map
			}
/*		}
		else
		{
			int genome = (int)genomeDist.pickNumber();

			float h = hDist.pickNumber();

			m_masses[genome] += h;

			if (m_masses[genome] < 0)
				m_masses[genome] = smallDist.pickNumber(); // avoid creating a totally zero mass map
		}*/
	}
}

void GridLensInversionGenomeNew::initGenome()
{
	for (int i = 0 ; i < m_masses.size() ; i++)
		m_masses[i] = 0.0f;

	int numMutations = m_masses.size() * 2;
	//int numMutations = m_masses.size()/2;

	double minHeight = 0;
	double maxHeight = 1;
	const std::vector<Vector2D<float> > &centers = m_pFactory->getBasisCenters();
	float minWidth = m_pFactory->getMinimalMutationWidth();
	float maxWidth = m_pFactory->getMaximalMutationWidth();
	Vector2D<float> topRight = m_pFactory->getTopRight();
	Vector2D<float> bottomLeft = m_pFactory->getBottomLeft();
	//GaussianDistribution widthDist(0, maxWidth/5.0, m_pFactory->getRandomNumberGenerator());
	PeakDistribution widthDist(minWidth, maxWidth, m_pFactory->getRandomNumberGenerator());
	UniformDistribution xDist(bottomLeft.getX(), topRight.getX(), m_pFactory->getRandomNumberGenerator());
	UniformDistribution yDist(bottomLeft.getY(), topRight.getY(), m_pFactory->getRandomNumberGenerator());
	//UniformDistribution hDist(minHeight, maxHeight, m_pFactory->getRandomNumberGenerator());
	//JeffreysDistribution hDist(0.0000001, 1000000.0,  m_pFactory->getRandomNumberGenerator());
	//PeakDistribution hDist(0.1, 1000000.0, m_pFactory->getRandomNumberGenerator());
	StudentTDistribution hDist(0.2, m_pFactory->getRandomNumberGenerator());
	//GaussianDistribution hDist(0, 0.2, m_pFactory->getRandomNumberGenerator());

	for (int i = 0 ; i < numMutations ; i++)
	{
		Vector2D<float> center((float)xDist.pickNumber(), (float)yDist.pickNumber());
		float width = widthDist.pickNumber()/3.0;
		float width2 = width*width;
		float h = ABS(hDist.pickNumber());

		for (int j = 0 ; j < centers.size() ; j++)
		{
			Vector2D<float> diff = centers[j]-center;
			float distSquared = diff.getLengthSquared();

			m_masses[j] += h*EXP(-distSquared/(width2*2.0));
			if (m_masses[j] < 0)
				m_masses[j] = 0;
		}
	}

	float maxValue = 0;
	float minValue = 0;

	for (int i = 0 ; i < m_masses.size() ; i++)
	{
		if (m_masses[i] > maxValue)
			maxValue = m_masses[i];
		if (m_masses[i] < minValue)
			minValue = m_masses[i];
	}

	float scaleFactor = MAX(ABS(minValue),ABS(maxValue));

	for (int i = 0 ; i < m_masses.size() ; i++)
		m_masses[i] /= scaleFactor;
}

std::string GridLensInversionGenomeNew::getFitnessDescription() const
{
	char str[1024];
	char *pStr = str;
	int numComponents = m_pFactory->getNumberOfFitnessComponents();

	for (int i = 0 ; i < numComponents ; i++)
	{
		sprintf(pStr,"%g ",(double)m_fitnessValues[i]);
		pStr += strlen(pStr);
	}
	return std::string(str);
}
	
GravitationalLens *GridLensInversionGenomeNew::createLens(double *totalmass, std::string &errstr) const
{
	return m_pFactory->createLens(m_masses, m_sheetValue, m_scaleFactor, totalmass, errstr);
}

void GridLensInversionGenomeNew::initializeNewCalculation()
{
	// TODO: this is probably not necessary
	for (int i = 0 ; i < m_masses.size() ; i++)
		m_masses[i] *= m_scaleFactor;
	m_scaleFactor = 1.0f;

	GridLensInversionGAFactoryNew *pFactory = getFactory();
	DeflectionMatrix *pDeflectionMatrix = pFactory->getDeflectionMatrix();
	BackProjectMatrixNew *pTotalBackProjectMatrix = pFactory->getTotalBackProjectMatrix();
	BackProjectMatrixNew *pShortBackProjectMatrix = pFactory->getShortBackProjectMatrix();

	pDeflectionMatrix->calculateBasisMatrixProducts(getMasses(), true, true, true);
	pTotalBackProjectMatrix->storeDeflectionMatrixResults();
	pShortBackProjectMatrix->storeDeflectionMatrixResults();
}

bool GridLensInversionGenomeNew::calculateMassScaleFitness(float scaleFactor, float sheetScale, float &fitness)
{
	BackProjectMatrixNew *pShortBackProjectMatrix = m_pFactory->getShortBackProjectMatrix();
	LensFitnessObject &fitnessFunction = *(m_pFactory->getFitnessObject());

	pShortBackProjectMatrix->calculate(scaleFactor, sheetScale);
	if (fitnessFunction.shortNeedInverseMagnifications())
		pShortBackProjectMatrix->calculateInverseMagnifications(*(fitnessFunction.getShortInverseMagnificationFlags()));
	if (fitnessFunction.shortNeedShearComponents())
		pShortBackProjectMatrix->calculateShearComponents(*(fitnessFunction.getShortShearComponentFlags()));
	if (fitnessFunction.shortNeedConvergence())
		pShortBackProjectMatrix->calculateConvergence(*(fitnessFunction.getShortConvergenceFlags()));

	if (!fitnessFunction.calculateMassScaleFitness(*pShortBackProjectMatrix, fitness))
	{
		cerr << "ERROR: Unable to calculate mass scale fitness: " << fitnessFunction.getErrorString() << endl;
		return false;
	}
	return true;
}

bool GridLensInversionGenomeNew::calculateTotalFitness(float scaleFactor, float sheetScale, float *pFitnessValues)
{
	GridLensInversionGAFactoryNew *m_pFactory = getFactory();
	BackProjectMatrixNew *pTotalBackProjectMatrix = m_pFactory->getTotalBackProjectMatrix();
	LensFitnessObject &fitnessFunction = *(m_pFactory->getFitnessObject());
	
	pTotalBackProjectMatrix->calculate(scaleFactor, sheetScale);
	if (fitnessFunction.totalNeedInverseMagnifications())
		pTotalBackProjectMatrix->calculateInverseMagnifications(*(fitnessFunction.getTotalInverseMagnificationFlags()));
	if (fitnessFunction.totalNeedShearComponents())
		pTotalBackProjectMatrix->calculateShearComponents(*(fitnessFunction.getTotalShearComponentFlags()));
	if (fitnessFunction.totalNeedConvergence())
		pTotalBackProjectMatrix->calculateConvergence(*(fitnessFunction.getTotalConvergenceFlags()));

	if (!fitnessFunction.calculateOverallFitness(*pTotalBackProjectMatrix, pFitnessValues))
	{
		cerr << "ERROR: Unable to calculate full fitness: " << fitnessFunction.getErrorString() << endl;
		return false;
	}
	return true;
}

} // end namespace


