/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#ifndef REALTYPE_H

#define REALTYPE_H

#include <string.h>
#include <cmath>
#ifdef GRALECONFIG_SUPPORT_INTELIPP
	#include <ipps.h>
#endif // GRALECONFIG_SUPPORT_INTELIPP

namespace grale
{
#ifndef WIN32
	inline float asinh(float x)
	{
		return ::asinhf(x);
	}
	
	inline double asinh(double x)
	{
		return ::asinh(x);
	}

	inline long double asinh(long double x)
	{
		return ::asinhl(x);
	}
	
	inline float acosh(float x)
	{
		return ::acoshf(x);
	}
	
	inline double acosh(double x)
	{
		return ::acosh(x);
	}

	inline long double acosh(long double x)
	{
		return ::acoshl(x);
	}	
	
	inline float atanh(float x)
	{
		return ::atanhf(x);
	}
	
	inline double atanh(double x)
	{
		return ::atanh(x);
	}

	inline long double atanh(long double x)
	{
		return ::atanhl(x);
	}

	template<class T>
	inline T min(T x, T y)
	{
		return (x < y)?x:y;
	}
	
	template<class T>
	inline T max(T x, T y)
	{
		return (x > y)?x:y;
	}
#else
	template<class T>
	T asinh(T x)
	{
		return (std::exp(x)-std::exp(-x))/(T)2;
	}
	
	template<class T>
	T acosh(T x)
	{
		return (std::exp(x)+std::exp(-x))/(T)2;
	}

	template<class T>
	T atanh(T x)
	{
		return (std::exp(x)-std::exp(-x))/(std::exp(x)+std::exp(-x));
	}
#endif

#ifndef GRALECONFIG_SUPPORT_INTELIPP

	template<class T>
	inline T CalculateDotProduct(const T *a, const T *b, int num)
	{
		T v = 0;
		for (int i = 0 ; i < num ; i++)
			v += a[i]*b[i];
		return v;
	}

	template<class T>
	inline void CalculateAddProductC(T *pDest, const T *pSrc, T C, int num, T *tmpbuf)
	{
		for (int i = 0 ; i < num ; i++)
			pDest[i] = pDest[i] + pSrc[i]*C;
	}

	template<class T>
	inline void CopyVector(T *pDest, const T *pSrc, int num)
	{
		memcpy(pDest, pSrc, num*sizeof(T));
	}

	template<class T>
	inline void SquareVector(T *pVect, int num)
	{
		for (int i = 0 ; i < num ; i++)
			pVect[i] *= pVect[i];
	}

	template<class T>
	inline void SquareVector(T *pDest, const T *pSrc, int num)
	{
		for (int i = 0 ; i < num ; i++)
			pDest[i] = pSrc[i]*pSrc[i];
	}
	template<class T>
	inline void MultiplyVector(T *pSrcDst, const T *pSrc, int num)
	{
		for (int i = 0 ; i < num ; i++)
			pSrcDst[i] *= pSrc[i];
	}

	template<class T>
	inline void MultiplyVector(T *pDst, const T *pSrc, const T factor, int num)
	{
		for (int i = 0 ; i < num ; i++)
			pDst[i] = pSrc[i]*factor;
	}

	template<class T>
	inline void MultiplyVector(T *pSrcDst, const T factor, int num)
	{
		for (int i = 0 ; i < num ; i++)
			pSrcDst[i] *= factor;
	}
	
	template<class T>
	inline void MultiplyVector(T *pDst, const T *pSrc1, const T *pSrc2, int num)
	{
		for (int i = 0 ; i < num ; i++)
			pDst[i] = pSrc1[i]*pSrc2[i];
	}

	template<class T>
	inline void AbsVector(T *pVect, int num)
	{
		for (int i = 0 ; i < num ; i++)
			pVect[i] = std::abs(pVect[i]);
	}

	template<class T>
	inline void AddVector(T *pSrcDst, const T *pSrc, int num)
	{
		for (int i = 0 ; i < num ; i++)
			pSrcDst[i] += pSrc[i];
	}

	template<class T>
	inline void AddVector(T *pSrcDst, T value, int num)
	{
		for (int i = 0 ; i < num ; i++)
			pSrcDst[i] += value;
	}

	template<class T>
	inline void AddVector(T *pDst, const T *pSrc1, const T *pSrc2, int num)
	{
		for (int i = 0 ; i < num ; i++)
			pDst[i] = pSrc1[i] + pSrc2[i];
	}

	template<class T>
	inline void SubVector(T *pSrcDst, const T *pSrc, int num)
	{
		for (int i = 0 ; i < num ; i++)
			pSrcDst[i] -= pSrc[i];
	}

	template<class T>
	inline void SubVector(T *pDst, const T *pSrc1, const T *pSrc2, int num)
	{
		for (int i = 0 ; i < num ; i++)
			pDst[i] = pSrc1[i] - pSrc2[i];
	}


#else

//#error "TODO: add float versions"

	inline double CalculateDotProduct(const double *a, const double *b, int num)
	{
		double val;
		ippsDotProd_64f(a, b, num, &val);
		return val;
	}

	inline void CalculateAddProductC(double *pDest, const double *pSrc, double C, int num, double *tmpbuf)
	{
		ippsMulC_64f(pSrc, C, tmpbuf, num);
		ippsAdd_64f_I(tmpbuf, pDest, num);
	}

	inline void CopyVector(double *pDest, const double *pSrc, int num)
	{
		ippsCopy_64f(pSrc, pDest, num);
	}

	inline void SquareVector(double *pVect, int num)
	{
		ippsSqr_64f_I(pVect, num);
	}

	inline void MultiplyVector(double *pSrcDst, const double *pSrc, int num)
	{
		ippsMul_64f_I(pSrc, pSrcDst, num);
	}

	inline void MultiplyVector(double *pSrcDst, double factor, int num)
	{
		ippsMulC_64f_I(factor, pSrcDst, num);
	}

	inline void AbsVector(double *pVect, int num)
	{
		ippsAbs_64f_I(pVect, num);
	}

	inline void AddVector(double *pSrcDst, const double *pSrc, int num)
	{
		ippsAdd_64f_I(pSrc, pSrcDst, num);
	}

	inline void SubVector(double *pSrcDst, const double *pSrc, int num)
	{
		ippsSub_64f_I(pSrc, pSrcDst, num);
	}

	inline void SubVector(double *pDst, const double *pSrc1, const double *pSrc2, int num)
	{
		ippsSub_64f(pSrc2, pSrc1, pDst, num);
	}

	// float versions

	inline float CalculateDotProduct(const float *a, const float *b, int num)
	{
		float val;
		ippsDotProd_32f(a, b, num, &val);
		return val;
	}

	inline void CalculateAddProductC(float *pDest, const float *pSrc, float C, int num, float *tmpbuf)
	{
		ippsMulC_32f(pSrc, C, tmpbuf, num);
		ippsAdd_32f_I(tmpbuf, pDest, num);
	}

	inline void CopyVector(float *pDest, const float *pSrc, int num)
	{
		ippsCopy_32f(pSrc, pDest, num);
	}

	inline void SquareVector(float *pVect, int num)
	{
		ippsSqr_32f_I(pVect, num);
	}

	inline void SquareVector(float *pDst, const float *pSrc, int num)
	{
		ippsSqr_32f(pSrc, pDst, num);
	}

	inline void MultiplyVector(float *pSrcDst, const float *pSrc, int num)
	{
		ippsMul_32f_I(pSrc, pSrcDst, num);
	}

	inline void MultiplyVector(float *pDst, const float *pSrc, float factor, int num)
	{
		ippsMulC_32f(pSrc, factor, pDst, num);
	}

	inline void MultiplyVector(float *pDst, const float *pSrc1, const float *pSrc2, int num)
	{
		ippsMul_32f(pSrc1, pSrc2, pDst, num);
	}

	inline void AbsVector(float *pVect, int num)
	{
		ippsAbs_32f_I(pVect, num);
	}

	inline void AddVector(float *pSrcDst, const float *pSrc, int num)
	{
		ippsAdd_32f_I(pSrc, pSrcDst, num);
	}

	inline void AddVector(float *pDst, const float *pSrc1, const float *pSrc2, int num)
	{
		ippsAdd_32f(pSrc1, pSrc2, pDst, num);
	}

	inline void AddVector(float *pSrcDst, float value, int num)
	{
		ippsAddC_32f_I(value, pSrcDst, num);
	}

	inline void SubVector(float *pSrcDst, const float *pSrc, int num)
	{
		ippsSub_32f_I(pSrc, pSrcDst, num);
	}

	inline void SubVector(float *pDst, const float *pSrc1, const float *pSrc2, int num)
	{
		ippsSub_32f(pSrc2, pSrc1, pDst, num);
	}

	inline void MultiplyVector(float *pSrcDst, float factor, int num)
	{
		ippsMulC_32f_I(factor, pSrcDst, num);
	}
	
#endif // !GRALECONFIG_SUPPORT_INTELIPP


}

#define ABS(x) 		std::abs(x)
#define SQRT(x)		std::sqrt(x)
#define COS(x)		std::cos(x)
#define SIN(x)		std::sin(x)
#define TAN(x)		std::tan(x)
#define ACOS(x)		std::acos(x)
#define ASIN(x)		std::asin(x)
#define ATAN(x)		std::atan(x)
#define LN(x)		std::log(x)
#define ASINH(x)	grale::asinh(x)
#define ACOSH(x)	grale::acosh(x)
#define ATANH(x)	grale::atanh(x)
#define POW(x,y)	std::pow(x,y)
#define EXP(x)		std::exp(x)
#define SINH(x)		std::sinh(x)
#define COSH(x)		std::cosh(x)
#define TANH(x)		std::tanh(x)

#ifndef WIN32
#define MIN(x,y)	grale::min(x,y)
#define MAX(x,y)	grale::max(x,y)
#else
#define MIN(x,y) (((x)<(y))?(x):(y))
#define MAX(x,y) (((x)>(y))?(x):(y))
#endif


#endif // REALTYPE_H

