/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "graleconfig.h" 
#include "imagesdata.h"
#include "polygon2d.h"
#include "triangle2d.h"
#include "rectangularcoordinates.h"
#include "wcscoordinates.h"
#include <serut/fileserializer.h>
#ifdef GRALECONFIG_SUPPORT_CFITSIO
	#include <fitsio.h>
#endif // GRALECONFIG_SUPPORT_CFITSIO
#ifdef GRALECONFIG_SUPPORT_QT5
	#include <QImage>
#endif // GRALECONFIG_SUPPORT_QT5
#include <list>
#include <map>

#include "debugnew.h"

namespace grale
{

ImagesData::ImagesData()
{
}

void ImagesData::copyFrom(const ImagesData &data)
{
	m_images = data.m_images;
	m_intensities = data.m_intensities;
	m_shearComponent1s = data.m_shearComponent1s;
	m_shearComponent2s = data.m_shearComponent2s;
	m_groupPoints = data.m_groupPoints;
	m_timeDelayInfo = data.m_timeDelayInfo;
	m_triangulations = data.m_triangulations;
	m_topRight = data.m_topRight;
	m_bottomLeft = data.m_bottomLeft;
}

ImagesData::~ImagesData()
{
	clear();
}

void ImagesData::clear()
{
	m_images.clear();
	m_intensities.clear();
	m_shearComponent1s.clear();
	m_shearComponent2s.clear();
	m_groupPoints.clear();
	m_triangulations.clear();
	m_timeDelayInfo.clear();
	m_bottomLeft = Vector2D<double>(0,0);
	m_topRight = Vector2D<double>(0,0);
}

ImagesData *ImagesData::createCopy() const
{
	ImagesData *imgdata = new ImagesData();

	imgdata->m_images = m_images;
	imgdata->m_intensities = m_intensities;
	imgdata->m_shearComponent2s = m_shearComponent2s;
	imgdata->m_shearComponent1s = m_shearComponent1s;
	imgdata->m_groupPoints = m_groupPoints;
	imgdata->m_triangulations = m_triangulations;
	imgdata->m_topRight = m_topRight;
	imgdata->m_bottomLeft = m_bottomLeft;

	return imgdata;
}

bool ImagesData::create(int numImages, bool intensities, bool shearInfo)
{
	if (numImages <= 0)
	{
		setErrorString("The number of images must be greater than zero");
		return false;
	}
		
	clear();
	
	m_images.resize(numImages);
	if (intensities)
		m_intensities.resize(numImages);
	if (shearInfo)
	{
		m_shearComponent1s.resize(numImages);
		m_shearComponent2s.resize(numImages);
	}
	return true;
}

int ImagesData::addImage()
{
	int newNumImages = m_images.size()+1;

	m_images.resize(newNumImages);
	if (m_intensities.size() > 0)
		m_intensities.resize(newNumImages);
	if (m_shearComponent1s.size() > 0)
	{
		m_shearComponent1s.resize(newNumImages);
		m_shearComponent2s.resize(newNumImages);
	}
	if (m_triangulations.size() > 0)
		m_triangulations.resize(newNumImages);

	return (newNumImages-1); // index of the new image
}
	
int ImagesData::addPoint(int imageNumber, Vector2D<double> point)
{
	if (imageNumber < 0 || imageNumber >= m_images.size())
	{
		setErrorString("Invalid image number");
		return -1;
	}
	
	if (m_intensities.size() > 0)
	{
		setErrorString("Intensity information must be specified for this images data set");
		return -1;
	}
	if (m_shearComponent1s.size() > 0)
	{
		setErrorString("Shear information must be specified for this images data set");
		return -1;
	}
	
	int pointIndex = m_images[imageNumber].size();
	
	m_images[imageNumber].resize(pointIndex+1);
	m_images[imageNumber][pointIndex] = point;
	
	findExtremes();
	
	return pointIndex;
}

int ImagesData::addPoint(int imageNumber, Vector2D<double> point, double intensity)
{
	if (imageNumber < 0 || imageNumber >= m_images.size())
	{
		setErrorString("Invalid image number");
		return false;
	}

	if (m_intensities.size() == 0)
	{
		setErrorString("No intensity information can be specified for this images data set");
		return false;
	}
	if (m_shearComponent1s.size() > 0)
	{
		setErrorString("Shear information must be specified for this images data set");
		return -1;
	}

	int pointIndex = m_images[imageNumber].size();
	
	m_images[imageNumber].resize(pointIndex+1);
	m_images[imageNumber][pointIndex] = point;
	m_intensities[imageNumber].resize(pointIndex+1);
	m_intensities[imageNumber][pointIndex] = intensity;
	
	findExtremes();

	return pointIndex;
}

int ImagesData::addPoint(int imageNumber, Vector2D<double> point, double shearComponent1, double shearComponent2)
{
	if (imageNumber < 0 || imageNumber >= m_images.size())
	{
		setErrorString("Invalid image number");
		return false;
	}

	if (m_intensities.size() > 0)
	{
		setErrorString("Intensity information must be specified for this images data set");
		return false;
	}
	if (m_shearComponent1s.size() == 0)
	{
		setErrorString("No shear information can be specified for this images data set");
		return -1;
	}

	int pointIndex = m_images[imageNumber].size();
	
	m_images[imageNumber].resize(pointIndex+1);
	m_images[imageNumber][pointIndex] = point;
	m_shearComponent1s[imageNumber].resize(pointIndex+1);
	m_shearComponent1s[imageNumber][pointIndex] = shearComponent1;
	m_shearComponent2s[imageNumber].resize(pointIndex+1);
	m_shearComponent2s[imageNumber][pointIndex] = shearComponent2;
	
	findExtremes();
	return pointIndex;
}

int ImagesData::addPoint(int imageNumber, Vector2D<double> point, double intensity, double shearComponent1, double shearComponent2)
{
	if (imageNumber < 0 || imageNumber >= m_images.size())
	{
		setErrorString("Invalid image number");
		return false;
	}

	if (m_intensities.size() == 0)
	{
		setErrorString("No intensity information can be specified for this images data set");
		return false;
	}
	if (m_shearComponent1s.size() == 0)
	{
		setErrorString("No shear information can be specified for this images data set");
		return -1;
	}

	int pointIndex = m_images[imageNumber].size();
	
	m_images[imageNumber].resize(pointIndex+1);
	m_images[imageNumber][pointIndex] = point;
	m_intensities[imageNumber].resize(pointIndex+1);
	m_intensities[imageNumber][pointIndex] = intensity;
	m_shearComponent1s[imageNumber].resize(pointIndex+1);
	m_shearComponent1s[imageNumber][pointIndex] = shearComponent1;
	m_shearComponent2s[imageNumber].resize(pointIndex+1);
	m_shearComponent2s[imageNumber][pointIndex] = shearComponent2;
	
	findExtremes();
	return pointIndex;
}

int ImagesData::addGroup()
{
	if (m_images.size() == 0)
	{
		setErrorString("No images exist yet");
		return -1;
	}

	int groupIndex = m_groupPoints.size();

	m_groupPoints.resize(groupIndex+1);
	
	return groupIndex;
}

bool ImagesData::addGroupPoint(int groupNumber, int imageIndex, int pointIndex)
{
	if (groupNumber < 0 || groupNumber >= m_groupPoints.size())
	{
		setErrorString("Invalid group number");
		return false;
	}

	if (imageIndex < 0 || imageIndex >= m_images.size())
	{
		setErrorString("Invalid image index");
		return false;
	}

	if (pointIndex < 0 || pointIndex >= m_images[imageIndex].size())
	{
		setErrorString("Invalid point index");
		return false;
	}
	
	int pos = m_groupPoints[groupNumber].size();
	
	m_groupPoints[groupNumber].resize(pos+2);
	m_groupPoints[groupNumber][pos] = imageIndex;
	m_groupPoints[groupNumber][pos+1] = pointIndex;
	
	return true;
}
	
bool ImagesData::addTimeDelayInfo(int imageIndex, int pointIndex, double timeDelay)
{
	if (imageIndex < 0 || imageIndex >= m_images.size())
	{
		setErrorString("Invalid image index");
		return false;
	}

	if (pointIndex < 0 || pointIndex >= m_images[imageIndex].size())
	{
		setErrorString("Invalid point index");
		return false;
	}

	m_timeDelayInfo.push_back(TimeDelayPoint(imageIndex, pointIndex, timeDelay));

	return true;
}

bool ImagesData::addTriangle(int imageIndex, int index1, int index2, int index3)
{
	if (m_triangulations.size() == 0)
		m_triangulations.resize(m_images.size());

	if (imageIndex < 0 || imageIndex >= m_images.size())
	{
		setErrorString("Invalid image index");
		return false;
	}

	if (index1 < 0 || index1 >= m_images[imageIndex].size())
	{
		setErrorString("Invalid point index 1");
		return false;
	}
	if (index2 < 0 || index2 >= m_images[imageIndex].size())
	{
		setErrorString("Invalid point index 2");
		return false;
	}
	if (index3 < 0 || index3 >= m_images[imageIndex].size())
	{
		setErrorString("Invalid point index 3");
		return false;
	}

	if (index1 == index2 || index2 == index3 || index1 == index3)
	{
		setErrorString("Two point indices are equal");
		return false;
	}

	m_triangulations[imageIndex].push_back(TriangleIndices(index1, index2, index3));

	return true;
}

bool ImagesData::createGridImage(Vector2D<double> bottomleft, Vector2D<double> topright, int numx, int numy, 
		                 const std::list<Triangle2D<double> > &excluderegions, bool triangulate,
				 bool useIntens)
{
	if (numx < 2 || numy < 2)
	{
		setErrorString("Number of points in X and Y directions must be at least 2");
		return false;
	}
	
	clear();

	if (!triangulate)
	{
		std::list<Vector2D<double> > points;
		std::list<Vector2D<double> > scaledpoints;

		double dx = (topright.getX()-bottomleft.getX())/((double)(numx-1));
		double dy = (topright.getY()-bottomleft.getY())/((double)(numy-1));
		double yscale = ABS(dy);
		double xscale = ABS(dx);
		double y = bottomleft.getY();
		double sy = bottomleft.getY()/yscale;
		
		for (int j = 0 ; j < numy ; j++, y += dy, sy += dy/yscale)
		{
			double x = bottomleft.getX();
			double sx = bottomleft.getX()/xscale;
			for (int i = 0 ; i < numx ; i++, x += dx, sx += dx/xscale)
			{
				points.push_back(Vector2D<double>(x,y));
				scaledpoints.push_back(Vector2D<double>(sx,sy));
			}
		}

		// recale the triangles

		std::vector<Triangle2D<double> > scaledtriangles(excluderegions.size());
		std::list<Triangle2D<double> >::const_iterator it;
		int idx;

		for (it = excluderegions.begin(), idx = 0 ; it != excluderegions.end() ; it++, idx++)
		{
			const Vector2D<double> *pPoints = (*it).getPoints();
			const Vector2D<double> p1(pPoints[0].getX()/xscale, pPoints[0].getY()/yscale);
			const Vector2D<double> p2(pPoints[1].getX()/xscale, pPoints[1].getY()/yscale);
			const Vector2D<double> p3(pPoints[2].getX()/xscale, pPoints[2].getY()/yscale);
			
			scaledtriangles[idx] = Triangle2D<double>(p1,p2,p3);
		}

		std::list<Vector2D<double> >::iterator i1 = points.begin();
		std::list<Vector2D<double> >::iterator i2 = scaledpoints.begin();

		while (i2 != scaledpoints.end())
		{
			bool hit = false;

			for (int i = 0 ; !hit && i < scaledtriangles.size() ; i++)
			{
				if (scaledtriangles[i].isInside(*i2))
					hit = true;
			}

			if (hit)
			{
				i1 = points.erase(i1);
				i2 = scaledpoints.erase(i2);
			}
			else
			{
				i1++;
				i2++;
			}
		}

		// store the remaining points

		m_images.resize(1);
		m_images[0].resize(points.size());

		if (useIntens)
		{
			m_intensities.resize(1);
			m_intensities[0].resize(points.size());
		}

		int i = 0;
		for (i1 = points.begin() ; i1 != points.end() ; i1++, i++)
		{
			m_images[0][i] = (*i1);
			if (useIntens)
				m_intensities[0][i] = 0;
		}
	}
	else
	{
		std::vector<Vector2D<double> > points(numx*numy);
		std::vector<Vector2D<double> > scaledpoints(numx*numy);
		std::vector<TriangleIndices> triangles((numx-1)*(numy-1)*2);

		double dx = (topright.getX()-bottomleft.getX())/((double)(numx-1));
		double dy = (topright.getY()-bottomleft.getY())/((double)(numy-1));
		double yscale = ABS(dy);
		double xscale = ABS(dx);
		double y = bottomleft.getY();
		double sy = bottomleft.getY()/yscale;
		int index = 0;
		
		for (int j = 0 ; j < numy ; j++, y += dy, sy += dy/yscale)
		{
			double x = bottomleft.getX();
			double sx = bottomleft.getX()/xscale;
			for (int i = 0 ; i < numx ; i++, x += dx, sx += dx/xscale, index++)
			{
				points[index] = Vector2D<double>(x,y);
				scaledpoints[index] = Vector2D<double>(sx,sy);
			}
		}

		index = 0;
		for (int j = 0 ; j < numy-1 ; j++)
		{
			for (int i = 0 ; i < numx-1 ; i++)
			{
				triangles[index++] = TriangleIndices(i + j*numx, i + 1 + j*numx, i + (j+1)*numx);
				triangles[index++] = TriangleIndices(i+1 + (j+1)*numx, i+1 + j*numx, i + (j+1)*numx);
			}
		}

		// rescale input triangles

		std::vector<Triangle2D<double> > scaledtriangles(excluderegions.size());
		std::list<Triangle2D<double> >::const_iterator it;
		int idx;

		for (it = excluderegions.begin(), idx = 0 ; it != excluderegions.end() ; it++, idx++)
		{
			const Vector2D<double> *pPoints = (*it).getPoints();
			const Vector2D<double> p1(pPoints[0].getX()/xscale, pPoints[0].getY()/yscale);
			const Vector2D<double> p2(pPoints[1].getX()/xscale, pPoints[1].getY()/yscale);
			const Vector2D<double> p3(pPoints[2].getX()/xscale, pPoints[2].getY()/yscale);
			
			scaledtriangles[idx] = Triangle2D<double>(p1,p2,p3);
		}

		// For each triangle, we'll check if there's overlap with an input triangle. If so, we'll ignore it

		std::list<TriangleIndices> acceptedTriangles;
		int numTriangles = (numx-1)*(numy-1)*2;
		int nextStartIdx = 0;

		for (int t = 0 ; t < numTriangles ; t++)
		{
			bool accept = true;
			Vector2D<double> p0 = scaledpoints[triangles[t].getIndex(0)];
			Vector2D<double> p1 = scaledpoints[triangles[t].getIndex(1)];
			Vector2D<double> p2 = scaledpoints[triangles[t].getIndex(2)];
			double minx0 = MIN(MIN(p0.getX(), p1.getX()), p2.getX());
			double miny0 = MIN(MIN(p0.getY(), p1.getY()), p2.getY());
			double maxx0 = MAX(MAX(p0.getX(), p1.getX()), p2.getX());
			double maxy0 = MAX(MAX(p0.getY(), p1.getY()), p2.getY());
			Triangle2D<double> curTriangle(p0, p1, p2);

			for (int i = 0 ; accept && i < scaledtriangles.size() ; i++)
			{
				int j = (i+nextStartIdx)%scaledtriangles.size();
				Vector2D<double> q0 = scaledtriangles[j].getPoints()[0];
				Vector2D<double> q1 = scaledtriangles[j].getPoints()[1];
				Vector2D<double> q2 = scaledtriangles[j].getPoints()[2];
				double minx1 = MIN(MIN(q0.getX(), q1.getX()), q2.getX());
				double miny1 = MIN(MIN(q0.getY(), q1.getY()), q2.getY());
				double maxx1 = MAX(MAX(q0.getX(), q1.getX()), q2.getX());
				double maxy1 = MAX(MAX(q0.getY(), q1.getY()), q2.getY());

				if (maxx0 < minx1 || minx0 > maxx1 || maxy0 < miny1 || miny0 > maxy1)
				{
					// can't overlap
				}
				else
				{
					double overlap = scaledtriangles[j].getOverlapArea(curTriangle);

					if (overlap > 0)
					{
						nextStartIdx = j;
						accept = false;
					}
				}
			}

			if (accept)
				acceptedTriangles.push_back(triangles[t]);
		}

		// Ok, now we've got all the triangles. Just need to store them properly.

		std::list<TriangleIndices>::const_iterator it2;

		m_images.resize(1);
		m_images[0].clear();
		m_triangulations.resize(1);
		m_triangulations[0].clear();

		if (useIntens)
			m_intensities.resize(1);

		if (useIntens)
		{
			for (it2 = acceptedTriangles.begin() ; it2 != acceptedTriangles.end() ; it2++)
			{
				int idx1 = processTrianglePoint(points[(*it2).getIndex(0)], m_images[0], m_intensities[0]);
				int idx2 = processTrianglePoint(points[(*it2).getIndex(1)], m_images[0], m_intensities[0]);
				int idx3 = processTrianglePoint(points[(*it2).getIndex(2)], m_images[0], m_intensities[0]);

				m_triangulations[0].push_back(TriangleIndices(idx1, idx2, idx3));
			}	
		}
		else
		{
			for (it2 = acceptedTriangles.begin() ; it2 != acceptedTriangles.end() ; it2++)
			{
				int idx1 = processTrianglePoint(points[(*it2).getIndex(0)], m_images[0]);
				int idx2 = processTrianglePoint(points[(*it2).getIndex(1)], m_images[0]);
				int idx3 = processTrianglePoint(points[(*it2).getIndex(2)], m_images[0]);

				m_triangulations[0].push_back(TriangleIndices(idx1, idx2, idx3));
			}	
		}

	}

	findExtremes();

	return true;
}

struct VectorOrder
{
	bool operator()(Vector2D<double> a, Vector2D<double> b) const
	{
		if (a.getX() < b.getX())
			return true;
		if (a.getX() > b.getX())
			return false;
		if (a.getY() < b.getY())
			return true;
		return false;
	}
};

bool ImagesData::createGridImage(Vector2D<double> bottomLeft, Vector2D<double> topRight, int numX, int numY, 
	                     const std::list<ImagesData *> &excludeImages, bool useIntens, bool storeMinimalDistances)
{

	if (numX < 2 || numY < 2)
	{
		setErrorString("Number of points in X and Y directions must be at least 2");
		return false;
	}

	if (useIntens && storeMinimalDistances)
	{
		setErrorString("Can't store intensities and minimal distances at the same time");
		return false;
	}
	
	double dx = (topRight.getX()-bottomLeft.getX())/(double)numX;
	double dy = (topRight.getY()-bottomLeft.getY())/(double)numY;
	double xScale = MAX(ABS(topRight.getX()), ABS(bottomLeft.getX()));
	double yScale = MAX(ABS(topRight.getY()), ABS(bottomLeft.getY()));
	double scaleFactor = SQRT(xScale*xScale+yScale*yScale);

	std::vector<Vector2D<double> > triangulationPoints;

	for (int i = 0 ; i < numX ; i++)
	{
		for (int j = 0 ; j < numY ; j++)
		{
			Vector2D<double> point = bottomLeft;

			point += Vector2D<double>(dx*(double)i, dy*(double)j);
			triangulationPoints.push_back(point/scaleFactor);
		}
	}

	std::list<ImagesData *>::const_iterator it;

	std::vector<std::pair<int, int> > constraints;
	std::vector<Polygon2D<double> > borders;

	for (it = excludeImages.begin() ; it != excludeImages.end() ; it++)
	{
		const ImagesData *pImgDat = *it;
		int numImages = pImgDat->getNumberOfImages();

		for (int i = 0 ; i < numImages ; i++)
		{
			Polygon2D<double> border;
	
			if (!pImgDat->getImageBorder(i, border))
			{
				setErrorString(std::string("Error obtaining border of an image: ") + pImgDat->getErrorString());
				return false;
			}

			if (storeMinimalDistances)
				borders.push_back(border);

			border.scale(1.01);

			int numPoints = border.getNumberOfPoints();
			const Vector2D<double> *pPoints = border.getPoints();

			int startIndex = triangulationPoints.size();

			for (int p = 0 ; p < numPoints ; p++)
				triangulationPoints.push_back(pPoints[p]/scaleFactor);
			
			int endIndex = triangulationPoints.size();

			for (int p = startIndex ; p < endIndex-1 ; p++)
				constraints.push_back(std::pair<int, int>(p, p+1));
			constraints.push_back(std::pair<int, int>(endIndex-1, startIndex));
		}
	}

	Triangulation triangulation;

	if (!triangulation.create(triangulationPoints, constraints))
	{
		setErrorString(std::string("Unable to create a constrained triangulation: ") + triangulation.getErrorString());
		return false;
	}
	
	std::list<TriangleIndices> nullTriangles;
	std::list<TriangleIndices>::const_iterator triangIt;
	std::list<Triangle2D<double> > finalTriangles;

	triangulation.getTriangles(nullTriangles);

	for (triangIt = nullTriangles.begin() ; triangIt != nullTriangles.end() ; triangIt++)
	{
		bool found = false;
		TriangleIndices triangleIndices = *triangIt;
		Vector2D<double> point1 = triangulationPoints[triangleIndices.getIndex(0)];
		Vector2D<double> point2 = triangulationPoints[triangleIndices.getIndex(1)];
		Vector2D<double> point3 = triangulationPoints[triangleIndices.getIndex(2)];
		Triangle2D<double> nullTriangle(point1, point2, point3);

		for (it = excludeImages.begin() ; !found && it != excludeImages.end() ; it++)
		{
			const ImagesData *pImgDat = *it;
			int numImages = pImgDat->getNumberOfImages();
	
			for (int i = 0 ; !found && i < numImages ; i++)
			{
				std::list<TriangleIndices> excludeTriangles;
				std::list<TriangleIndices>::const_iterator it2;

				pImgDat->getTriangles(i, excludeTriangles);

				for (it2 = excludeTriangles.begin() ; !found && it2 != excludeTriangles.end() ; it2++)
				{
					TriangleIndices excludeIndices = *it2;

					Vector2D<double> point4 = pImgDat->getImagePointPosition(i, excludeIndices.getIndex(0))/scaleFactor;
					Vector2D<double> point5 = pImgDat->getImagePointPosition(i, excludeIndices.getIndex(1))/scaleFactor;
					Vector2D<double> point6 = pImgDat->getImagePointPosition(i, excludeIndices.getIndex(2))/scaleFactor;

					Triangle2D<double> excludeTriangle(point4, point5, point6);

					if (nullTriangle.getOverlapArea(excludeTriangle) > 0)
						found = true;
				}
			}
		}

		if (!found) // ok, save this null space triangle
			finalTriangles.push_back(nullTriangle);
	}

	// Got the final set of triangles, now create new image with it

	clear();
	if (!create(1, useIntens||storeMinimalDistances, false))
		return false;

	std::list<Triangle2D<double> >::const_iterator it2;
	std::map<Vector2D<double>, int, VectorOrder> pointMap;
	std::map<Vector2D<double>, int, VectorOrder>::iterator it3;

	for (it2 = finalTriangles.begin() ; it2 != finalTriangles.end() ; it2++)
	{
		Triangle2D<double> t = *it2;
		int indices[3];

		for (int i = 0 ; i < 3 ; i++)
		{
			Vector2D<double> p = t.getPoints()[i];

			it3 = pointMap.find(p);
			if (it3 != pointMap.end())
				indices[i] = (*it3).second;
			else
			{
				if (useIntens||storeMinimalDistances)
					indices[i] = addPoint(0, p*scaleFactor, 0);
				else
					indices[i] = addPoint(0, p*scaleFactor);

				if (indices[i] < 0)
					return false;

				pointMap[p] = indices[i];
			}
		}
	
		if (!addTriangle(0, indices[0], indices[1], indices[2]))
			return false;
	}

	if (storeMinimalDistances)
	{
		for (int p = 0 ; p < m_images[0].size() ; p++)
			m_intensities[0][p] = -1.0;

		for (int b = 0 ; b < borders.size() ; b++)
		{
			for (int p = 0 ; p < m_images[0].size() ; p++)
			{
				double curValue = m_intensities[0][p];
				double distanceSquared = borders[b].getDistanceSquared(m_images[0][p]);

				if (curValue < 0 || distanceSquared < curValue)
					m_intensities[0][p] = distanceSquared;
			}
		}
	}
	return true;
}

bool ImagesData::addGridImage(Vector2D<double> bottomleft, Vector2D<double> topright, int numx, int numy, 
		                 const std::list<Triangle2D<double> > &excluderegions, bool triangulate)
{
	if (numx < 2 || numy < 2)
	{
		setErrorString("Number of points in X and Y directions must be at least 2");
		return false;
	}
	
	if (!triangulate)
	{
		std::list<Vector2D<double> > points;
		std::list<Vector2D<double> > scaledpoints;

		double dx = (topright.getX()-bottomleft.getX())/((double)(numx-1));
		double dy = (topright.getY()-bottomleft.getY())/((double)(numy-1));
		double yscale = ABS(dy);
		double xscale = ABS(dx);
		double y = bottomleft.getY();
		double sy = bottomleft.getY()/yscale;
		
		for (int j = 0 ; j < numy ; j++, y += dy, sy += dy/yscale)
		{
			double x = bottomleft.getX();
			double sx = bottomleft.getX()/xscale;
			for (int i = 0 ; i < numx ; i++, x += dx, sx += dx/xscale)
			{
				points.push_back(Vector2D<double>(x,y));
				scaledpoints.push_back(Vector2D<double>(sx,sy));
			}
		}

		// recale the triangles

		std::vector<Triangle2D<double> > scaledtriangles(excluderegions.size());
		std::list<Triangle2D<double> >::const_iterator it;
		int idx;

		for (it = excluderegions.begin(), idx = 0 ; it != excluderegions.end() ; it++, idx++)
		{
			const Vector2D<double> *pPoints = (*it).getPoints();
			const Vector2D<double> p1(pPoints[0].getX()/xscale, pPoints[0].getY()/yscale);
			const Vector2D<double> p2(pPoints[1].getX()/xscale, pPoints[1].getY()/yscale);
			const Vector2D<double> p3(pPoints[2].getX()/xscale, pPoints[2].getY()/yscale);
			
			scaledtriangles[idx] = Triangle2D<double>(p1,p2,p3);
		}

		std::list<Vector2D<double> >::iterator i1 = points.begin();
		std::list<Vector2D<double> >::iterator i2 = scaledpoints.begin();

		while (i2 != scaledpoints.end())
		{
			bool hit = false;

			for (int i = 0 ; !hit && i < scaledtriangles.size() ; i++)
			{
				if (scaledtriangles[i].isInside(*i2))
					hit = true;
			}

			if (hit)
			{
				i1 = points.erase(i1);
				i2 = scaledpoints.erase(i2);
			}
			else
			{
				i1++;
				i2++;
			}
		}

		// store the remaining points

		int newIndex = addImage();
		bool hasIntens = hasIntensities();

		int i = 0;
		for (i1 = points.begin() ; i1 != points.end() ; i1++, i++)
		{
			if (hasIntens)
				addPoint(newIndex, *i1, 0);
			else
				addPoint(newIndex, *i1);
		}
	}
	else
	{
		std::vector<Vector2D<double> > points(numx*numy);
		std::vector<Vector2D<double> > scaledpoints(numx*numy);
		std::vector<TriangleIndices> triangles((numx-1)*(numy-1)*2);

		double dx = (topright.getX()-bottomleft.getX())/((double)(numx-1));
		double dy = (topright.getY()-bottomleft.getY())/((double)(numy-1));
		double yscale = ABS(dy);
		double xscale = ABS(dx);
		double y = bottomleft.getY();
		double sy = bottomleft.getY()/yscale;
		int index = 0;
		
		for (int j = 0 ; j < numy ; j++, y += dy, sy += dy/yscale)
		{
			double x = bottomleft.getX();
			double sx = bottomleft.getX()/xscale;
			for (int i = 0 ; i < numx ; i++, x += dx, sx += dx/xscale, index++)
			{
				points[index] = Vector2D<double>(x,y);
				scaledpoints[index] = Vector2D<double>(sx,sy);
			}
		}

		index = 0;
		for (int j = 0 ; j < numy-1 ; j++)
		{
			for (int i = 0 ; i < numx-1 ; i++)
			{
				triangles[index++] = TriangleIndices(i + j*numx, i + 1 + j*numx, i + (j+1)*numx);
				triangles[index++] = TriangleIndices(i+1 + (j+1)*numx, i+1 + j*numx, i + (j+1)*numx);
			}
		}

		// rescale input triangles

		std::vector<Triangle2D<double> > scaledtriangles(excluderegions.size());
		std::list<Triangle2D<double> >::const_iterator it;
		int idx;

		for (it = excluderegions.begin(), idx = 0 ; it != excluderegions.end() ; it++, idx++)
		{
			const Vector2D<double> *pPoints = (*it).getPoints();
			const Vector2D<double> p1(pPoints[0].getX()/xscale, pPoints[0].getY()/yscale);
			const Vector2D<double> p2(pPoints[1].getX()/xscale, pPoints[1].getY()/yscale);
			const Vector2D<double> p3(pPoints[2].getX()/xscale, pPoints[2].getY()/yscale);
			
			scaledtriangles[idx] = Triangle2D<double>(p1,p2,p3);
		}

		// For each triangle, we'll check if there's overlap with an input triangle. If so, we'll ignore it

		std::list<TriangleIndices> acceptedTriangles;
		int numTriangles = (numx-1)*(numy-1)*2;
		int nextStartIdx = 0;

		for (int t = 0 ; t < numTriangles ; t++)
		{
			bool accept = true;
			Vector2D<double> p0 = scaledpoints[triangles[t].getIndex(0)];
			Vector2D<double> p1 = scaledpoints[triangles[t].getIndex(1)];
			Vector2D<double> p2 = scaledpoints[triangles[t].getIndex(2)];
			double minx0 = MIN(MIN(p0.getX(), p1.getX()), p2.getX());
			double miny0 = MIN(MIN(p0.getY(), p1.getY()), p2.getY());
			double maxx0 = MAX(MAX(p0.getX(), p1.getX()), p2.getX());
			double maxy0 = MAX(MAX(p0.getY(), p1.getY()), p2.getY());
			Triangle2D<double> curTriangle(p0, p1, p2);

			for (int i = 0 ; accept && i < scaledtriangles.size() ; i++)
			{
				int j = (i+nextStartIdx)%scaledtriangles.size();
				Vector2D<double> q0 = scaledtriangles[j].getPoints()[0];
				Vector2D<double> q1 = scaledtriangles[j].getPoints()[1];
				Vector2D<double> q2 = scaledtriangles[j].getPoints()[2];
				double minx1 = MIN(MIN(q0.getX(), q1.getX()), q2.getX());
				double miny1 = MIN(MIN(q0.getY(), q1.getY()), q2.getY());
				double maxx1 = MAX(MAX(q0.getX(), q1.getX()), q2.getX());
				double maxy1 = MAX(MAX(q0.getY(), q1.getY()), q2.getY());

				if (maxx0 < minx1 || minx0 > maxx1 || maxy0 < miny1 || miny0 > maxy1)
				{
					// can't overlap
				}
				else
				{
					double overlap = scaledtriangles[j].getOverlapArea(curTriangle);

					if (overlap > 0)
					{
						nextStartIdx = j;
						accept = false;
					}
				}
			}

			if (accept)
				acceptedTriangles.push_back(triangles[t]);
		}

		// Ok, now we've got all the triangles. Just need to store them properly.

		std::list<TriangleIndices>::const_iterator it2;

		int newIndex = addImage();

		m_triangulations.resize(newIndex+1); 
		m_triangulations[newIndex].clear();

		if (hasIntensities())
		{
			for (it2 = acceptedTriangles.begin() ; it2 != acceptedTriangles.end() ; it2++)
			{
				int idx1 = processTrianglePoint(points[(*it2).getIndex(0)], m_images[newIndex], m_intensities[newIndex]);
				int idx2 = processTrianglePoint(points[(*it2).getIndex(1)], m_images[newIndex], m_intensities[newIndex]);
				int idx3 = processTrianglePoint(points[(*it2).getIndex(2)], m_images[newIndex], m_intensities[newIndex]);

				m_triangulations[newIndex].push_back(TriangleIndices(idx1, idx2, idx3));
			}	
		}
		else
		{
			for (it2 = acceptedTriangles.begin() ; it2 != acceptedTriangles.end() ; it2++)
			{
				int idx1 = processTrianglePoint(points[(*it2).getIndex(0)], m_images[newIndex]);
				int idx2 = processTrianglePoint(points[(*it2).getIndex(1)], m_images[newIndex]);
				int idx3 = processTrianglePoint(points[(*it2).getIndex(2)], m_images[newIndex]);

				m_triangulations[newIndex].push_back(TriangleIndices(idx1, idx2, idx3));
			}	
		}
	}

	findExtremes();

	return true;
}

#define IMAGESDATAID 0x41544449

bool ImagesData::load(const std::string &fname)
{
	serut::FileSerializer fser;
	
	if (!fser.open(fname, serut::FileSerializer::ReadOnly))
	{
		setErrorString(std::string("Can't load images data: ") + fser.getErrorString());
		return false;
	}
	if (!read(fser))
		return false;
	return true;
}

bool ImagesData::save(const std::string &fname) const
{
	serut::FileSerializer fser;

	if (!fser.open(fname, serut::FileSerializer::WriteOnly))
	{
		setErrorString(std::string("Can't save images data: ") + fser.getErrorString());
		return false;
	}
	if (!write(fser))
		return false;
	return true;
}

#define IMAGESDATA_FLAG_INTENSITIES	1
#define IMAGESDATA_FLAG_TRIANGULATIONS	2
#define IMAGESDATA_FLAG_TIMEDELAY	4 
#define IMAGESDATA_FLAG_SHEAR		8

bool ImagesData::read(serut::SerializationInterface &si)
{
	int32_t id, numimgs, flag;
	bool gotintens, gottriang, gotTimeDelay, gotShear;
	std::vector<std::vector<Vector2D<double> > > imgs;
	std::vector<std::vector<double> > intens, shearComponent1s, shearComponent2s;
	std::vector<TimeDelayPoint> timeDelays;
	std::vector<int32_t> numpoints;
	
	if (!si.readInt32(&id))
	{
		setErrorString(std::string("Error reading images data ID: ") + si.getErrorString());
		return false;
	}
	if (id != IMAGESDATAID)
	{
		setErrorString("Read invalid images data ID");
		return false;
	}
	if (!si.readInt32(&flag))
	{
		setErrorString(std::string("Error reading feature flag: ") + si.getErrorString());
		return false;
	}

	gotintens = ((flag&IMAGESDATA_FLAG_INTENSITIES) == 0)?false:true;
	gottriang = ((flag&IMAGESDATA_FLAG_TRIANGULATIONS) == 0)?false:true;
	gotTimeDelay = ((flag&IMAGESDATA_FLAG_TIMEDELAY) == 0)?false:true;
	gotShear = ((flag&IMAGESDATA_FLAG_SHEAR) == 0)?false:true;
	
	if (!si.readInt32(&numimgs))
	{
		setErrorString(std::string("Error reading number of images: ") + si.getErrorString());
		return false;
	}

	imgs.resize(numimgs);
	numpoints.resize(numimgs);
	if (gotintens)
		intens.resize(numimgs);
	if (gotShear)
	{
		shearComponent1s.resize(numimgs);
		shearComponent2s.resize(numimgs);
	}
	
	if (!si.readInt32s(numpoints))
	{
		setErrorString(std::string("Error reading images: ") + si.getErrorString());
		return false;
	}

	for (int i = 0 ; i < numimgs ; i++)
	{
		int num = numpoints[i];

		imgs[i].resize(num);
		if (gotintens)
			intens[i].resize(num);
		if (gotShear)
		{
			shearComponent1s[i].resize(num);
			shearComponent2s[i].resize(num);
		}
	}

	for (int i = 0 ; i < numimgs ; i++)
	{
		int num = numpoints[i];
		bool err = false;

		for (int j = 0 ; !err && j < imgs[i].size() ; j++)
		{
			if (!si.readDoubles(imgs[i][j].getComponents(), 2))
				err = true;
		}

		if (!err)
		{
			if (gotintens)
			{
				if (!si.readDoubles(intens[i]))
					err = true;
			}
			if (gotShear && !err)
			{
				if (!si.readDoubles(shearComponent1s[i]))
					err = true;
				else if (!si.readDoubles(shearComponent2s[i]))
					err = true;
			}
		}

		if (err)
		{
			setErrorString(std::string("Error reading images: ") + si.getErrorString());
			return false;
		}
	}

	int32_t groups;

	if (!si.readInt32(&groups))
	{
		setErrorString(std::string("Error reading number of groups: ") + si.getErrorString());
		return false;
	}

	std::vector<int32_t> numgrppts;
	std::vector<std::vector<int32_t> > grppts;

	if (groups > 0)
	{
		numgrppts.resize(groups);
		grppts.resize(groups);

		bool err = false;

		if (!si.readInt32s(numgrppts))
			err = true;

		for (int i = 0 ; !err && i < groups ; i++)
		{
			int num = numgrppts[i]*2;

			grppts[i].resize(num);
			if (!si.readInt32s(grppts[i]))
				err = true;
		}

		if (err)
		{
			setErrorString(std::string("Error reading group info: ") + si.getErrorString());
			return false;
		}
	}

	std::vector<std::list<TriangleIndices> > triangulations;
	
	if (gottriang)
	{
		triangulations.resize(imgs.size());

		for (int i = 0 ; i < imgs.size() ; i++)
		{
			int32_t num;

			if (!si.readInt32(&num))
			{
				setErrorString(std::string("Error reading number of triangles: ") + si.getErrorString());
				return false;
			}

			for (int32_t j = 0 ; j < num ; j++)
			{
				int32_t indices[3];

				if (!si.readInt32s(indices, 3))
				{
					setErrorString(std::string("Error reading triangle indices: ") + si.getErrorString());
					return false;
				}

				triangulations[i].push_back(TriangleIndices(indices[0], indices[1], indices[2]));
			}
		}
	}

	if (gotTimeDelay)
	{
		int32_t numDelays;

		if (!si.readInt32(&numDelays))
		{
			setErrorString(std::string("Error reading number of time delay points: ") + si.getErrorString());
			return false;
		}

		for (int32_t i = 0 ; i < numDelays ; i++)
		{
			int32_t indices[2];
			double delay;

			if (!si.readInt32s(indices, 2))
			{
				setErrorString(std::string("Error reading time delay point indices: ") + si.getErrorString());
				return false;
			}
			if (!si.readDouble(&delay))
			{
				setErrorString(std::string("Error reading time delay: ") + si.getErrorString());
				return false;
			}
			timeDelays.push_back(TimeDelayPoint(indices[0], indices[1], delay));
		}
	}

	clear(); // clear old data

	// save new data

	m_images = imgs;
	m_intensities = intens;
	m_groupPoints = grppts;
	m_triangulations = triangulations;
	m_timeDelayInfo = timeDelays;
	m_shearComponent1s = shearComponent1s;
	m_shearComponent2s = shearComponent2s;

	findExtremes();

	return true;
}

bool ImagesData::write(serut::SerializationInterface &si) const
{
	int flag = 0;

	if (m_intensities.size() != 0)
		flag |= IMAGESDATA_FLAG_INTENSITIES;
	if (m_triangulations.size() != 0)
		flag |= IMAGESDATA_FLAG_TRIANGULATIONS;
	if (m_timeDelayInfo.size() != 0)
		flag |= IMAGESDATA_FLAG_TIMEDELAY;
	if (m_shearComponent1s.size() != 0)
		flag |= IMAGESDATA_FLAG_SHEAR;

	if (!si.writeInt32(IMAGESDATAID))
	{
		setErrorString(std::string("Error writing images data ID: ") + si.getErrorString());
		return false;
	}
	if (!si.writeInt32(flag))
	{
		setErrorString(std::string("Error writing feature flag: ") + si.getErrorString());
		return false;
	}
	if (!si.writeInt32(m_images.size()))
	{
		setErrorString(std::string("Error writing number of images: ") + si.getErrorString());
		return false;
	}

	std::vector<int32_t> numImagePoints(m_images.size());

	for (int i = 0 ; i < numImagePoints.size() ; i++)
		numImagePoints[i] = m_images[i].size();

	if (!si.writeInt32s(numImagePoints))
	{
		setErrorString(std::string("Error writing images: ") + si.getErrorString());
		return false;
	}

	for (int i = 0 ; i < m_images.size() ; i++)
	{
		for (int j = 0 ; j < m_images[i].size() ; j++)
		{
			if (!si.writeDoubles(m_images[i][j].getComponents(), 2))
			{
				setErrorString(std::string("Error writing images: ") + si.getErrorString());
				return false;
			}
		}
		if (m_intensities.size() != 0)
		{
			if (!si.writeDoubles(m_intensities[i]))
			{
				setErrorString(std::string("Error writing images: ") + si.getErrorString());
				return false;
			}
		}
		if (m_shearComponent1s.size() != 0)
		{
			if (!si.writeDoubles(m_shearComponent1s[i]))
			{
				setErrorString(std::string("Error writing images: ") + si.getErrorString());
				return false;
			}
			if (!si.writeDoubles(m_shearComponent2s[i]))
			{
				setErrorString(std::string("Error writing images: ") + si.getErrorString());
				return false;
			}
		}
	}

	if (!si.writeInt32(m_groupPoints.size()))
	{
		setErrorString(std::string("Error writing number of groups: ") + si.getErrorString());
		return false;
	}

	if (m_groupPoints.size() > 0)
	{
		std::vector<int32_t> numgrouppoints(m_groupPoints.size());

		for (int i = 0 ; i < m_groupPoints.size() ; i++)
			numgrouppoints[i] = m_groupPoints[i].size()/2;

		if (!si.writeInt32s(numgrouppoints))
		{
			setErrorString(std::string("Error writing number of group points: ") + si.getErrorString());
			return false;
		}

		for (int i = 0 ; i < m_groupPoints.size() ; i++)
		{
			if (!si.writeInt32s(m_groupPoints[i]))
			{
				setErrorString(std::string("Error writing group points: ") + si.getErrorString());
				return false;
			}
		}
	}

	if (m_triangulations.size() != 0)
	{
		for (int i = 0 ; i < m_images.size() ; i++) // this is the amount of triangulations we should have
		{
			int32_t indices[3];
			
			// write size

			if (!si.writeInt32(m_triangulations[i].size()))
			{
				setErrorString(std::string("Error writing triangulation size: ") + si.getErrorString());
				return false;
			}

			std::list<TriangleIndices>::const_iterator it;

			for (it = m_triangulations[i].begin() ; it != m_triangulations[i].end() ; it++)
			{
				indices[0] = (*it).getIndex(0);
				indices[1] = (*it).getIndex(1);
				indices[2] = (*it).getIndex(2);

				if (!si.writeInt32s(indices, 3))
				{
					setErrorString(std::string("Error writing triangulation indices: ") + si.getErrorString());
					return false;
				}
			}
		}
	}

	if (m_timeDelayInfo.size() != 0)
	{
		if (!si.writeInt32(m_timeDelayInfo.size()))
		{
			setErrorString(std::string("Error writing number of time delay points: ") + si.getErrorString());
			return false;
		}

		for (int i = 0 ; i < m_timeDelayInfo.size() ; i++)
		{
			int32_t indices[2];

			indices[0] = m_timeDelayInfo[i].getImageIndex();
			indices[1] = m_timeDelayInfo[i].getPointIndex();

			if (!si.writeInt32s(indices, 2))
			{
				setErrorString(std::string("Error writing time delay point indices: ") + si.getErrorString());
				return false;
			}

			if (!si.writeDouble(m_timeDelayInfo[i].getTimeDelay()))
			{
				setErrorString(std::string("Error writing time delay: ") + si.getErrorString());
				return false;
			}
		}
	}

	return true;
}

bool ImagesData::loadPNG(const std::string &fname, Vector2D<double> bottomleft, Vector2D<double> topright, bool nullspace, const std::string &maskfile, bool triangulate)
{
#ifdef GRALECONFIG_SUPPORT_QT5
	QImage pngimage;
	int numx, numy;
	
	if (!pngimage.load(fname.c_str()))
	{
		setErrorString("Error reading from PNG file");
		return false;
	}

	numx = pngimage.width();
	numy = pngimage.height();

	int numtotal = numx*numy;
	double *values = new double[numtotal];

	int x,y,i;

	i = 0;
	for (y = 0 ; y < numy ; y++)
	{
		for (x = 0 ; x < numx ; x++,i++)
		{
			QRgb color = pngimage.pixel(x, numy-1-y);
			int gray = qGray(color);

			values[i] = (double)gray/255.0;
		}
	}

	RectangularCoordinates rcoords(bottomleft,topright,numx,numy);
	bool status = processData(values, rcoords, numx, numy, false, nullspace, maskfile, triangulate);

	delete [] values;
	return status;
#else
	setErrorString("No PNG support was compiled in");
	return false;
#endif // GRALECONFIG_SUPPORT_QT5
}

bool ImagesData::loadFITS(const std::string &fname,bool nullspace,const std::string &maskfile, bool triangulate)
{
#ifdef GRALECONFIG_SUPPORT_CFITSIO
	fitsfile *f;
	long naxes[2];
	int nfound;
	int status = 0;

	if (fits_open_file(&f,fname.c_str(),READONLY,&status))
	{
		setErrorString("Can't open FITS file");
		return false;
	}

	bool axesfound = false;
	
	while(!axesfound)
	{
		bool err = false;
		
		if (fits_read_keys_lng(f, "NAXIS", 1, 2, naxes, &nfound, &status))
			err = true;
		else if (nfound != 2)
			err = true;

		if (err) // try next HDU
		{
			int t;
			
			if (fits_movrel_hdu(f, 1, &t, &status))
			{
				fits_close_file(f,&status);
				setErrorString("Can't read image dimensions");
				return false;
			}
		}
		else
			axesfound = true;
	}

	double *values = new double[naxes[0]*naxes[1]];
	double nullval = 0;
	int anynull = 0;
	
	if (fits_read_img(f,TDOUBLE,1,(int)(naxes[0]*naxes[1]),&nullval,values,&anynull,&status))
	{
		delete [] values;
		fits_close_file(f,&status);
		setErrorString("Can't read FITS image data");
		return false;
	}

	char *str = 0;
	int nkeys = 0;
	if (fits_hdr2str(f,0,0,0,&str,&nkeys,&status))
	{
		delete [] values;
		fits_close_file(f,&status);
		setErrorString("Can't read FITS header string");
		return false;
	}
	
	std::string headerstr(str);
	free(str);
	
	fits_close_file(f,&status);

	int numx = naxes[0];
	int numy = naxes[1];
	int numtotal = numx*numy;

	// Process the image itself

	WCSCoordinates wcscoords;

	if (!wcscoords.init(headerstr))
	{
		delete [] values;
		setErrorString(std::string("Unable to create a WCS coordinate converter: ") + wcscoords.getErrorString());
		return false;
	}
	
	bool ret = processData(values,wcscoords,numx,numy,true,nullspace,maskfile,triangulate);
	
	delete [] values;
	
	return ret;
#else
	setErrorString("No CFITSIO support was compiled in");
	return false;
#endif // GRALECONFIG_SUPPORT_CFITSIO
}

bool ImagesData::loadFITS(const std::string &fname,double unit,bool nullspace,const std::string &maskfile, bool triangulate)
{
#ifdef GRALECONFIG_SUPPORT_CFITSIO
	fitsfile *f;
	long naxes[2];
	int nfound;
	int status = 0;

	if (fits_open_file(&f,fname.c_str(),READONLY,&status))
	{
		setErrorString("Can't open FITS file");
		return false;
	}
	
	bool axesfound = false;
	
	while(!axesfound)
	{
		bool err = false;
		
		if (fits_read_keys_lng(f, "NAXIS", 1, 2, naxes, &nfound, &status))
			err = true;
		else if (nfound != 2)
			err = true;

		if (err) // try next HDU
		{
			int t;
			
			if (fits_movrel_hdu(f, 1, &t, &status))
			{
				fits_close_file(f,&status);
				setErrorString("Can't read image dimensions");
				return false;
			}
		}
		else
			axesfound = true;
	}

	double stepx;
	double stepy;
	double startx;
	double starty;
	char commstr[16384];
	
	if (fits_read_key_dbl(f,"CRVAL1",&startx,commstr,&status))
	{
		fits_close_file(f,&status);
		setErrorString("Can't read FITS CRVAL1 key");
		return false;
	}
	if (fits_read_key_dbl(f,"CRVAL2",&starty,commstr,&status))
	{
		fits_close_file(f,&status);
		setErrorString("Can't read FITS CRVAL2 key");
		return false;
	}	
	if (fits_read_key_dbl(f,"CDELT1",&stepx,commstr,&status))
	{
		fits_close_file(f,&status);
		setErrorString("Can't read FITS CDELT1 key");
		return false;
	}
	if (fits_read_key_dbl(f,"CDELT2",&stepy,commstr,&status))
	{
		fits_close_file(f,&status);
		setErrorString("Can't read FITS CDELT2 key");
		return false;
	}
	
	double *values = new double[naxes[0]*naxes[1]];
	double nullval = 0;
	int anynull = 0;
	
	if (fits_read_img(f,TDOUBLE,1,(int)(naxes[0]*naxes[1]),&nullval,values,&anynull,&status))
	{
		delete [] values;
		fits_close_file(f,&status);
		setErrorString("Can't read FITS image data");
		return false;
	}

	fits_close_file(f,&status);

	int numx = naxes[0];
	int numy = naxes[1];
	int numtotal = numx*numy;

	double dx = ((double)stepx)*unit;
	double dy = ((double)stepy)*unit;
	Vector2D<double> bottomleft(((double)startx)*unit,((double)starty)*unit);
	Vector2D<double> topright = bottomleft+Vector2D<double>(dx*(double)numx,dy*(double)numy);

	// Process the image itself

	RectangularCoordinates rcoords(bottomleft,topright,numx,numy);
	bool ret = processData(values,rcoords,numx,numy,true,nullspace,maskfile,triangulate);
	
	delete [] values;
	
	return ret;
#else
	setErrorString("No CFITSIO support was compiled in");
	return false;
#endif // GRALECONFIG_SUPPORT_CFITSIO
}

bool ImagesData::loadFITS(const std::string &fname,Vector2D<double> bottomleft,Vector2D<double> topright,bool nullspace,const std::string &maskfile,bool triangulate)
{
#ifdef GRALECONFIG_SUPPORT_CFITSIO
	fitsfile *f;
	long naxes[2];
	int nfound;
	int status = 0;

	if (fits_open_file(&f,fname.c_str(),READONLY,&status))
	{
		setErrorString("Can't open FITS file");
		return false;
	}

	bool axesfound = false;
	
	while(!axesfound)
	{
		bool err = false;
		
		if (fits_read_keys_lng(f, "NAXIS", 1, 2, naxes, &nfound, &status))
			err = true;
		else if (nfound != 2)
			err = true;

		if (err) // try next HDU
		{
			int t;
			
			if (fits_movrel_hdu(f, 1, &t, &status))
			{
				fits_close_file(f,&status);
				setErrorString("Can't read image dimensions");
				return false;
			}
		}
		else
			axesfound = true;
	}

	double *values = new double[naxes[0]*naxes[1]];
	double nullval = 0;
	int anynull = 0;
	
	if (fits_read_img(f,TDOUBLE,1,(int)(naxes[0]*naxes[1]),&nullval,values,&anynull,&status))
	{
		delete [] values;
		fits_close_file(f,&status);
		setErrorString("Can't read FITS image data");
		return false;
	}

	fits_close_file(f,&status);

	int numx = naxes[0];
	int numy = naxes[1];
	int numtotal = numx*numy;

	// Process the image itself

	RectangularCoordinates rcoords(bottomleft,topright,numx,numy);
	bool ret = processData(values,rcoords,numx,numy,true,nullspace,maskfile,triangulate);
	
	delete [] values;
	
	return ret;
#else
	setErrorString("No CFITSIO support was compiled in");
	return false;
#endif // GRALECONFIG_SUPPORT_CFITSIO
}


// We make sure that the null space images are processed in the same order as the images
// themselves

bool ImagesData::processData(double *values,CoordinateInterface &coordcalc,int numx,int numy,bool intens,bool nullspace,const std::string &maskfile,bool triangulate)
{
	int total = numx*numy;
	int i;
	std::vector<uint16_t> map(total);
	
	if (!createMap(values,numx,numy,maskfile,map))
		return false;
	
	std::list<std::list<IntVector2DPlus> > sourceImages;
	std::list<std::list<TriangleIndices> > triangulations;

	if (nullspace)
		triangulate = false; // can't automatically triangulate this
	
	if (!nullspace)
	{
		bool done = false;
	
		while (!done)
		{
			bool found = false;
			std::list<IntVector2DPlus> imagepositions;
	
			for (int i = 0 ; !found && i < total ; i++)
			{
				if (map[i] != 0)
				{
					found = true;
	
					int x = i%numx;
					int y = i/numx;
	
					imagepositions.push_back(IntVector2DPlus(x,y,map[i]));
					map[i] = 0;
				}
			}
	
			if (!found)
				done = true;
			else
			{
				bool done2 = false;
	
				while (!done2)
				{
					std::list<IntVector2DPlus> newpoints;
					std::list<IntVector2DPlus>::const_iterator it;
					bool found = false;
	
					for (it = imagepositions.begin() ; it != imagepositions.end() ; it++)
					{
						int x,y,index;
	
						x = (*it).getX();
						y = (*it).getY();
						index = x+y*numx;
#define SEARCHWIDTH 2
						for (int i = -SEARCHWIDTH ; i <= SEARCHWIDTH ; i++)
						{
							for (int j = -SEARCHWIDTH ; j <= SEARCHWIDTH ; j++)
							{
								int x2 = x+i;
								int y2 = y+j;
								int index2 = index+i+numx*j;
	
								if (x2 >= 0 && y2 >= 0 && x2 < numx && y2 < numy)
								{
									if (map[index2] != 0)
									{
										newpoints.push_back(IntVector2DPlus(x2,y2,map[index2]));
										map[index2] = 0;
										found = true;
									}
								}
							}
						}
					}
	
					if (!found)
						done2 = true;
					else
					{
						for (it = newpoints.begin() ; it != newpoints.end() ; it++)
							imagepositions.push_back(*it);
					}
				}
	
				// store this image
				
				sourceImages.push_back(imagepositions);

				if (triangulate)
				{
					std::list<IntLine2D> lines;
					std::list<TriangleIndices> triangles;
					
					createLineSegments(imagepositions, lines);
					createTriangulation(lines, triangles, numx);
					
					triangulations.push_back(triangles);
				}
			}
		}
	}
	else // null space
	{
		std::vector<bool> map2(total);
	
		for (i = 0 ; i < total ; i++)
		{
			int x = i%numx;
			int y = i/numy;

			if (map[i] == 0)
			{
				if ((x > 0 && map[i-1] != 0) ||
				    (x < numx-1 && map[i+1] != 0) ||
				    (y > 0 && map[i-numx] != 0) ||
				    (y < numy-1 && map[i+numx] != 0))
					map2[i] = true;
				else
					map2[i] = false;
			}
			else
				map2[i] = false;
		}
		
		bool done = false;
	
		while (!done)
		{
			bool found = false;
			std::list<IntVector2D> imagepositions;
			std::list<IntVector2D> borderpositions;
	
			for (int i = 0 ; !found && i < total ; i++)
			{
				if (map[i] != 0)
				{
					found = true;
	
					int x = i%numx;
					int y = i/numx;
	
					map[i] = 0;
					
					imagepositions.push_back(IntVector2D(x,y));

					if (x > 0 && map2[i-1])
					{
						map2[i-1] = false;
						borderpositions.push_back(IntVector2D(x-1,y));
					}
					if (x < numx-1 && map2[i+1])
					{
						map2[i+1] = false;
						borderpositions.push_back(IntVector2D(x+1,y));
					}
					if (y > 0 && map2[i-numx])
					{
						map2[i-numx] = false;
						borderpositions.push_back(IntVector2D(x,y-1));
					}
					if (y < numy-1 && map2[i+numx])
					{
						map2[i+numx] = false;
						borderpositions.push_back(IntVector2D(x,y+1));
					}

				}
			}
	
			if (!found)
				done = true;
			else
			{
				bool done2 = false;
	
				while (!done2)
				{
					std::list<IntVector2D> newpoints;
					std::list<IntVector2D>::const_iterator it;
					bool found = false;
	
					for (it = imagepositions.begin() ; it != imagepositions.end() ; it++)
					{
						int x,y,index;
	
						x = (*it).getX();
						y = (*it).getY();
						index = x+y*numx;
	
						for (int i = -SEARCHWIDTH ; i <= SEARCHWIDTH ; i++)
						{
							for (int j = -SEARCHWIDTH ; j <= SEARCHWIDTH ; j++)
							{
								int x2 = x+i;
								int y2 = y+j;
								int index2 = index+i+numx*j;
	
								if (x2 >= 0 && y2 >= 0 && x2 < numx && y2 < numy)
								{
									if (map[index2] != 0)
									{
										map[index2] = 0;
										newpoints.push_back(IntVector2D(x2,y2));
										found = true;

										if (x2 > 0 && map2[index2-1])
										{
											map2[index2-1] = false;
											borderpositions.push_back(IntVector2D(x2-1,y2));
										}
										if (x2 < numx-1 && map2[index2+1])
										{
											map2[index2+1] = false;
											borderpositions.push_back(IntVector2D(x2+1,y2));
										}
										if (y2 > 0 && map2[index2-numx])
										{
											map2[index2-numx] = false;
											borderpositions.push_back(IntVector2D(x2,y2-1));
										}
										if (y2 < numy-1 && map2[index2+numx])
										{
											map2[index2+numx] = false;
											borderpositions.push_back(IntVector2D(x2,y2+1));
										}
									}
								}
							}
						}
					}
	
					if (!found)
						done2 = true;
					else
					{
						for (it = newpoints.begin() ; it != newpoints.end() ; it++)
							imagepositions.push_back(*it);
					}
				}
	
				// store the actual coordinates in a list
				
				std::list<IntVector2DPlus> points;
				std::list<IntVector2D>::const_iterator it;
				
				for (it = borderpositions.begin() ; it != borderpositions.end() ; it++)
				{
					int x = (*it).getX();
					int y = (*it).getY();
	
					points.push_back(IntVector2DPlus(x,y,0xffff));
				}
	
				sourceImages.push_back(points);
			}
		}
	}
	
	// The sourceimages list now contains the data of the images in pixel coordinates and the indices of
	// points which should be matched.
		
	clear();

	if (sourceImages.empty())
		return true;

	int numimages = sourceImages.size();
	m_images.resize(numimages);

	if (intens)
		m_intensities.resize(numimages);
	if (triangulate)
		m_triangulations.resize(numimages);

	std::list<std::list<IntVector2DPlus> >::iterator imgit;
	std::list<std::list<TriangleIndices> >::const_iterator triangulationIt = triangulations.begin();

	for (i = 0, imgit = sourceImages.begin() ; i < numimages ; i++, imgit++)
	{
		std::map<int, int> triangleIndexMap;
		std::list<IntVector2DPlus> &imgpoints = (*imgit);
		std::list<IntVector2DPlus>::const_iterator it;
		int num = imgpoints.size();
		int j;
		
		m_images[i].resize(num);
		if (intens)
			m_intensities[i].resize(num);

		for (j = 0, it = imgpoints.begin() ; j < num ; j++, it++)
		{
			Vector2D<double> pos = coordcalc(*it);
			
			m_images[i][j] = pos;
			
			if (intens)
			{
				int x = (*it).getX();
				int y = (*it).getY();
				m_intensities[i][j] = values[x+y*numx];
			}
			
			if (triangulate)
			{
				int mapIndex = (*it).getX() + (*it).getY()*numx;

				triangleIndexMap[mapIndex] = j;
			}
		}

		if (triangulate)
		{
			std::list<TriangleIndices>::const_iterator it;

			for (it = (*triangulationIt).begin() ; it != (*triangulationIt).end() ; it++)
			{
				TriangleIndices indices = (*it);
				int mapIndex1 = indices.getIndex(0);
				int mapIndex2 = indices.getIndex(1);
				int mapIndex3 = indices.getIndex(2);
				int pointIndex1 = triangleIndexMap[mapIndex1];
				int pointIndex2 = triangleIndexMap[mapIndex2];
				int pointIndex3 = triangleIndexMap[mapIndex3];

				m_triangulations[i].push_back(TriangleIndices(pointIndex1, pointIndex2, pointIndex3));
			}

			triangulationIt++;
		}
	}

	findExtremes();

	return true;
}
	
void ImagesData::findExtremes()
{
	double minx,miny,maxx,maxy;
	bool extset = false;

	minx = 0;
	miny = 0;
	maxx = 0;
	maxy = 0;

	for (int img = 0 ; img < m_images.size() ; img++)
	{
		for (int imgpoint = 0 ; imgpoint < m_images[img].size() ; imgpoint++)
		{
			Vector2D<double> pos = getImagePointPosition(img, imgpoint);

			if (!extset)
			{
				extset = true;
				minx = pos.getX();
				maxx = minx;
				miny = pos.getY();
				maxy = miny;
			}
			else
			{
				if (pos.getX() < minx) minx = pos.getX();
				if (pos.getY() < miny) miny = pos.getY();
				if (pos.getX() > maxx) maxx = pos.getX();
				if (pos.getY() > maxy) maxy = pos.getY();
			}
		}
	}

	m_topRight = Vector2D<double>(maxx,maxy);
	m_bottomLeft = Vector2D<double>(minx,miny);
}

bool ImagesData::createMap(double *values, int numx, int numy, const std::string &maskfile, std::vector<uint16_t> &map)
{
	int total = numx*numy;

	if (maskfile.size() == 0)
	{
		for (int i = 0 ; i < total ; i++)
		{
			if (values[i] > 0)
				map[i] = 0xffff;
			else
				map[i] = 0;
		}
	}
	else // got a mask file!
	{
#ifdef GRALECONFIG_SUPPORT_QT5
		QImage pngimage;
		int nx,ny;
	
		if (!pngimage.load(maskfile.c_str()))
		{
			setErrorString("Error reading from PNG mask file");
			return false;
		}
	
		nx = pngimage.width();
		ny = pngimage.height();

		if ((numx != nx) || (numy != ny))
		{
			setErrorString("Incompatible mask file");
			return false;
		}

		int x,y;

		for (y = 0 ; y < numy ; y++)
		{
			for (x = 0 ; x < numx ; x++)
			{
				QRgb color = pngimage.pixel(x, numy-1-y);
				int R = qRed(color);
				int G = qGreen(color);
				int B = qBlue(color);
				int idx = x+y*numx;

				if (R == 0 && G == 0 && B == 0)
					map[idx] = 0;
				else
					map[idx] = 0xffff;
			}
		}
#else
		setErrorString("PNG support is required to be able to use mask files");
		return false;
#endif // GRALECONFIG_SUPPORT_QT5
	}
	return true;
}

// TODO!!!!
// Add this as 'CenterOnPosition(double ra, double dec'
void ImagesData::subtractPosition(Vector2D<double> pos)
{
	double a = pos.getX();
	double d = pos.getY();

	for (int img = 0 ; img < m_images.size() ; img++)
	{
		for (int imgpoint = 0 ; imgpoint < m_images[img].size() ; imgpoint++)
		{
			Vector2D<double> pos = m_images[img][imgpoint];
			
			double A = pos.getX();
			double D = pos.getY();
			
			double x2 = COS(d)*COS(D)*COS(A-a)+SIN(D)*SIN(d);
			double y2 = COS(D)*SIN(A-a);
			double z2 = -COS(D)*SIN(d)*COS(A-a)+SIN(D)*COS(d);

			double D2 = ASIN(z2);

			double xproj = x2/COS(D2);
			double yproj = y2/COS(D2);

			double A2 = ACOS(xproj);
			if (yproj < 0)
				A2 = -A2;
			
			m_images[img][imgpoint] = Vector2D<double>(A2,D2);
		}
	}
}

void ImagesData::subtractIntensity(double v)
{
	if (!hasIntensities())
		return;
	
	for (int img = 0 ; img < m_images.size() ; img++)
		for (int imgpoint = 0 ; imgpoint < m_images[img].size() ; imgpoint++)
			m_intensities[img][imgpoint] -= v;
}

bool ImagesData::createTriangulation()
{
	if (m_triangulations.size() != 0)
	{
		setErrorString("A triangulation is already present");
		return false;
	}
	
	for (int i = 0 ; i < m_images.size() ; i++)
	{
		if (m_images[i].size() < 3)
		{
			setErrorString("Each image in the images data set must contain at least three points");
			return false;
		}
	}

	std::vector<std::list<TriangleIndices> > triangulations(m_images.size());

	for (int i = 0 ; i < m_images.size() ; i++)
	{
		int numPoints = m_images[i].size();
		std::vector<Vector2D<double> > points(numPoints);

		for (int j = 0 ; j  < numPoints ; j++)
			points[j] = m_images[i][j];

		Triangulation triangulation;

		if (!triangulation.create(points))
		{
			setErrorString(std::string("Unable to create a triangulation: ") + triangulation.getErrorString());
			return false;
		}

		std::list<TriangleIndices> triangleIndices;
		std::list<TriangleIndices>::const_iterator it;

		triangulation.getTriangles(triangleIndices);

		// check which triangles are too flat
		// TODO: remove this? 

		for (it = triangleIndices.begin() ; it != triangleIndices.end() ; it++)
		{
			TriangleIndices indices = (*it);
			Vector2D<double> p1 = points[indices.getIndex(0)];
			Vector2D<double> p2 = points[indices.getIndex(1)];
			Vector2D<double> p3 = points[indices.getIndex(2)];
			Vector2D<double> diff1 = p2-p1;
			Vector2D<double> diff2 = p3-p2;
			Vector2D<double> diff3 = p1-p3;
			double d1 = diff1.getLength();
			double d2 = diff2.getLength();
			double d3 = diff3.getLength();
			grale::Vector2D<double> n1 = diff1/d1;
			grale::Vector2D<double> n2 = diff2/d2;
			Vector2D<double> n3 = diff3/d3;

			if (!(ABS(n1*n2) > 0.99999 || ABS(n2*n3) > 0.99999 || ABS(n3*n1) > 0.99999))
			{
				triangulations[i].push_back(indices);
			}
			else
				std::cerr << "Warning: found a triangle which is extremely flat" << std::endl;
		}
	}	

	m_triangulations = triangulations;

	return true;
}

bool ImagesData::clearTriangulation()
{
	if (m_triangulations.size() == 0)
	{
		setErrorString("No triangulations have been created");
		return false;
	}
	
	m_triangulations.clear();
	return true;
}

bool ImagesData::getTriangles(int image, std::list<TriangleIndices> &triangles) const
{
	if (m_triangulations.size() == 0)
	{
		setErrorString("No triangulations have been created");
		return false;
	}

	triangles = m_triangulations[image];
	return true;
}

int ImagesData::processTrianglePoint(Vector2D<double> p, std::vector<Vector2D<double> > &points)
{
	int num = points.size();

	for (int i = 0 ; i < num ; i++)
	{
		if (p.getX() == points[i].getX() && p.getY() == points[i].getY())
			return i;
	}
	points.push_back(p);
	return num;
}

int ImagesData::processTrianglePoint(Vector2D<double> p, std::vector<Vector2D<double> > &points, std::vector<double> &intensities)
{
	int num = points.size();

	for (int i = 0 ; i < num ; i++)
	{
		if (p.getX() == points[i].getX() && p.getY() == points[i].getY())
			return i;
	}
	points.push_back(p);
	intensities.push_back(0);
	return num;
}

void ImagesData::createLineSegments(const std::list<IntVector2DPlus> &points, std::list<IntLine2D> &lines)
{
	std::list<IntVector2DPlus>::const_iterator it, it2;
	
	lines.clear();
	
	for (it = points.begin() ; it != points.end() ; it++)
	{
		IntVector2D p((*it).getX(), (*it).getY());

		it2 = it;
		it2++;

		for (   ; it2 != points.end() ; it2++)
		{
			IntVector2D q((*it2).getX(), (*it2).getY());
			int dx = q.getX() - p.getX();
			int dy = q.getY() - p.getY();
			
			if (dx < 0)
				dx = -dx;
			if (dy < 0)
				dy = -dy;

			if ( (dx + dy) > 0  // avoid dx == 0 and dy == 0
					&& dx <= 1 && dy <= 1)
			{
				addLineSegment(lines, IntLine2D(p, q));
			}
		}
	}
}

void ImagesData::addLineSegment(std::list<IntLine2D> &lines, IntLine2D l)
{
	std::list<IntLine2D>::const_iterator it;
	IntVector2D p1 = l.getVertex1();
	IntVector2D p2 = l.getVertex2();
	int dx = p1.getX() - p2.getX();
	int dy = p1.getY() - p2.getY();
	int minx = MIN(p1.getX(), p2.getX());
	int miny = MIN(p1.getY(), p2.getY());
	int maxx = MAX(p1.getX(), p2.getX());
	int maxy = MAX(p1.getY(), p2.getY());

	if (dx < 0)
		dx = -dx;
	if (dy < 0)
		dy = -dy;

	bool checkCross = false;

	if (dx == 1 && dy == 1)
		checkCross = true;

	for (it = lines.begin() ; it != lines.end() ; it++)
	{
		IntLine2D curLine = (*it);
		IntVector2D q1 = curLine.getVertex1();
		IntVector2D q2 = curLine.getVertex2();
		
		// check if the lines are the same
		
		if (q1 == p1 && q2 == p2)
			return;
		if (q1 == p2 && q2 == p1)
			return;

		// check if the lines cross
		
		if (checkCross)
		{
			int dX = q1.getX() - q2.getX();
			int dY = q1.getY() - q2.getY();

			if (dX < 0)
				dX = -dX;
			if (dY < 0)
				dY = -dY;

			if (dX == 1 && dY == 1)
			{
				int minX = MIN(q1.getX(), q2.getX());
				int minY = MIN(q1.getY(), q2.getY());
				int maxX = MAX(q1.getX(), q2.getX());
				int maxY = MAX(q1.getY(), q2.getY());
					
				if (minx == minX && miny == minY && maxx == maxX && maxy == maxY) // crossing segments
					return;
			}
		}
	}

	// now we can safely add the line segment
	
	lines.push_back(l);
}

void ImagesData::createTriangulation(const std::list<IntLine2D> &lines, std::list<TriangleIndices> &triangles, int numx)
{
	std::list<IntLine2D>::const_iterator it1, it2, it3;

	for (it1 = lines.begin() ; it1 != lines.end() ; it1++)
	{
		IntVector2D p1 = (*it1).getVertex1();
		IntVector2D q1 = (*it1).getVertex2();

		it2 = it1;
		it2++;

		for (   ; it2 != lines.end() ; it2++)
		{
			IntVector2D p2 = (*it2).getVertex1();
			IntVector2D q2 = (*it2).getVertex2();
			IntVector2D a;
			IntVector2D b;
			IntVector2D c;
			bool connected = false;

			if (p1 == p2)
			{
				connected = true;
				a = q1;
				b = q2;
				c = p1;
			}
			else if (p1 == q2)
			{
				connected = true;
				a = q1;
				b = p2;
				c = p1;
			}
			else if (q1 == p2)
			{
				connected = true;
				a = p1;
				b = q2;
				c = q1;
			}
			else if (q1 == q2)
			{
				connected = true;
				a = p1;
				b = p2;
				c = q1;
			}

			int dx = a.getX() - b.getX();
			int dy = a.getY() - b.getY();

			if (dx < 0)
				dx = -dx;
			if (dy < 0)
				dy = -dy;

			if (connected && dx <= 1 && dy <= 1)
			{
				bool found = false;
				
				it3 = it2;
				it3++;
				
				for (   ; !found && it3 != lines.end() ; it3++)
				{
					IntVector2D p3 = (*it3).getVertex1();
					IntVector2D q3 = (*it3).getVertex2();

					if ((p3 == a && q3 == b) || (p3 == b && q3 == a))
						found = true;
				}

				if (found) // triangle a,b,c is present
				{
					int idx1 = a.getX() + a.getY()*numx;
					int idx2 = b.getX() + b.getY()*numx;
					int idx3 = c.getX() + c.getY()*numx;

					triangles.push_back(TriangleIndices(idx1, idx2, idx3));
				}
			}
		}
	}
}

struct SegmentOrder
{
	bool operator()(std::pair<int, int> x, std::pair<int, int> y) const
	{
		if (x.first < y.first)
			return true;
		if (x.first > y.first)
			return false;
		if (x.second < y.second)
			return true;
		return false;
	}
};

bool ImagesData::getImageBorder(int image, Polygon2D<double> &polygon) const
{
	if (!hasTriangulation())
	{
		setErrorString("No triangulation is present in this images data set");
		return false;
	}
	if (image < 0 || image >= m_images.size())
	{
		setErrorString("Invalid image number");
		return false;
	}

	std::list<TriangleIndices>::const_iterator it;
	std::map<std::pair<int, int>, int, SegmentOrder> segmentCounter;
	std::map<std::pair<int, int>, int, SegmentOrder>::iterator mapIt;

	for (it = m_triangulations[image].begin() ; it != m_triangulations[image].end() ; it++)
	{
		TriangleIndices triangle = *it;

		int idx1 = triangle.getIndex(0);
		int idx2 = triangle.getIndex(1);
		int idx3 = triangle.getIndex(2);

		if (idx1 == idx2 || idx2 == idx3 || idx1 == idx3)
		{
			setErrorString("Image contains an invalid triangulation");
			return false;
		}

		std::pair<int, int> segment1(MIN(idx1, idx2), MAX(idx1, idx2));
		std::pair<int, int> segment2(MIN(idx1, idx3), MAX(idx1, idx3));
		std::pair<int, int> segment3(MIN(idx2, idx3), MAX(idx2, idx3));

		mapIt = segmentCounter.find(segment1);
		if (mapIt != segmentCounter.end())
			(*mapIt).second++;
		else
			segmentCounter[segment1] = 1;

		mapIt = segmentCounter.find(segment2);
		if (mapIt != segmentCounter.end())
			(*mapIt).second++;
		else
			segmentCounter[segment2] = 1;

		mapIt = segmentCounter.find(segment3);
		if (mapIt != segmentCounter.end())
			(*mapIt).second++;
		else
			segmentCounter[segment3] = 1;
	}

	std::list<int> pointIndices;
	std::list<int>::const_iterator pointIt;
	std::list<std::pair<int, int> > outerEdges;

	for (mapIt = segmentCounter.begin() ; mapIt != segmentCounter.end() ; mapIt++)
	{
		if ((*mapIt).second == 1) // outer edge
			outerEdges.push_back((*mapIt).first);
	}

	bool done = false;
	while (!done)
	{
		int firstIndex = -1;
		int lastIndex = -2;

		if (!pointIndices.empty())
		{
			firstIndex = *(pointIndices.begin());

			pointIt = pointIndices.end();
			pointIt--;

			lastIndex = *pointIt;
		}

		if (firstIndex == lastIndex) // closed the loop
			done = true;
		else
		{
			bool found = false;

			std::list<std::pair<int, int> >::iterator segIt = outerEdges.begin();

			while (!found && segIt != outerEdges.end())
			{
				std::pair<int, int> edge = *segIt;

				if (pointIndices.empty())
				{
					found = true;

					pointIndices.push_back(edge.first);
					pointIndices.push_back(edge.second);
				}
				else
				{
					if (edge.first == firstIndex)
					{
						found = true;

						pointIndices.push_front(edge.second);
					}
					else if (edge.first == lastIndex)
					{
						found = true;

						pointIndices.push_back(edge.second);
					}
					else if (edge.second == firstIndex)
					{
						found = true;

						pointIndices.push_front(edge.first);
					}
					else if (edge.second == lastIndex)
					{
						found = true;

						pointIndices.push_back(edge.first);
					}
				}

				if (found)
					segIt = outerEdges.erase(segIt);
				else
					segIt++;
			}

			if (!found)
				done = true;
		}
	}

	if (!outerEdges.empty())
	{
		setErrorString("No single closed outer contour could be found");
		return false;
	}

	if (pointIndices.empty())
	{
		setErrorString("No image border could be found");
		return false;
	}

	int firstIndex = *(pointIndices.begin());

	pointIt = pointIndices.end();
	pointIt--;

	int lastIndex = *pointIt;

	if (firstIndex != lastIndex)
	{
		setErrorString("Found border is not closed");
		return false;
	}

	// Okay, create the polygon

	pointIt = pointIndices.begin();
	pointIt++; // first point is same as last, skip it

	std::list<Vector2D<double> > polyPoints;

	while (pointIt != pointIndices.end())
	{
		polyPoints.push_back(m_images[image][*pointIt]);
		pointIt++;
	}

	Polygon2D<double> imageBorder;

	imageBorder.init(polyPoints, false);
	polygon = imageBorder;

	return true;
}

} // end namespace

