/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "graleconfig.h"
#include "monopolegrid2.h"
#include "polygon2d.h"
#include "imagesdata.h"
#include "polynomialmassprofilelens.h"
#include <iostream>

#include "debugnew.h"

namespace grale
{

MonopoleGrid2::MonopoleGrid2()
{
}

MonopoleGrid2::~MonopoleGrid2()
{
}

bool MonopoleGrid2::init(const ImagesData &avoidImages, const ImagesData &deflectionPointImages,
		        const Real2DFunction &target, const GravitationalLens &differenceDeflection,
	                Vector2D<double> center, double areaSize, int subDiv, double zeroPointFactor)
{
	clear();

	// Create convex hulls for the images

	std::vector<Polygon2D<double> > convexHulls;

	for (int j = 0 ; j < avoidImages.getNumberOfImages() ; j++)
	{
		std::list<Vector2D<double> > points;

		for (int k = 0 ; k < avoidImages.getNumberOfImagePoints(j) ; k++)
			points.push_back(avoidImages.getImagePointPosition(j, k));
						
		Polygon2D<double> poly;

		poly.init(points, true);
		convexHulls.push_back(poly);
	}

	// Obtain the points at which the deflections need to be evaluated

	std::vector<Vector2D<double> > deflectionPoints;
	
	for (int j = 0 ; j < deflectionPointImages.getNumberOfImages() ; j++)
	{
		for (int k = 0 ; k < deflectionPointImages.getNumberOfImagePoints(j) ; k++)
			deflectionPoints.push_back(deflectionPointImages.getImagePointPosition(j, k));
	}

	// place the basis functions

	m_bottomLeft = center-Vector2D<double>(areaSize, areaSize)/2.0;
	m_stepSize = areaSize/(subDiv-1);

	for (int u = 0 ; u < subDiv ; u++)
	{
		for (int v = 0 ; v < subDiv ; v++)
		{
			Vector2D<double> center = m_bottomLeft + Vector2D<double>(u, v)*m_stepSize;
			double overallMinPos = 1e30;
			bool fixedMinPos = false;

			for (int i = 0 ; !fixedMinPos && i < convexHulls.size()  ; i++)
			{
				double minPos = 1e30;

				if (convexHulls[i].isInside(center))
				{
					overallMinPos = 0;
					fixedMinPos = true;
				}
				else
				{
					minPos = SQRT(convexHulls[i].getDistanceSquared(center));
					overallMinPos = MIN(overallMinPos, minPos);
				}
			}

			if (overallMinPos >= 4.0*m_stepSize) // give us some room
			{
				// check if it overlaps with an image point, if not, there's no point in adding
				// the monopole since it won't affect the deflection angle at any of the image
				// points
				
				double length = overallMinPos*0.99; // just a precaution
				double zeroPoint = (m_stepSize*zeroPointFactor)/length;
				bool found = false;

				for (int p = 0 ; !found && p < deflectionPoints.size() ; p++)
				{
					Vector2D<double> diff = deflectionPoints[p] - center;

					if (diff.getLength() < length)
						found = true;
				}
				
				if (found)
				{
					m_monoPoleFunctions.push_back(MonopoleBasisFunction(zeroPoint, length));
					m_monoPoleWeights.push_back(1.0);
					m_monoPoleCenters.push_back(center);
				}
			}
		}
	}
	
	// Got all the possible basis functions, now build the pixel dependencies so that we don't
	// have to iterate over all the basis functions when calculating the total effect of the
	// basis functions.
	// In the same loop, we'll also store the discrete version of the target.
	
	m_pixelDependencies.resize(subDiv*subDiv);
	m_targetMap.resize(subDiv*subDiv);
	
	int index = 0;
	int numBasisFunctions = m_monoPoleCenters.size();
	
	for (int y = 0 ; y < subDiv ; y++)
	{
		for (int x = 0 ; x < subDiv ; x++, index++)
		{
			Vector2D<double> pixelCenter = m_bottomLeft + Vector2D<double>(x, y)*m_stepSize;

			for (int b = 0 ; b < numBasisFunctions ; b++)
			{
				Vector2D<double> basisCenter = m_monoPoleCenters[b];
			
				Vector2D<double> diff = pixelCenter-basisCenter;

				if (diff.getLength() < m_monoPoleFunctions[b].getLengthScale())
				{
					// In this case, the basis function has an effect at the
					// pixel position
					
					m_pixelDependencies[index].push_back(b);
				}
			}
			
			m_targetMap[index] = target(pixelCenter);
		}
	}

	m_monoPoleMap.resize(subDiv*subDiv);
		
	m_subDiv = subDiv;

	m_differenceDeflections.resize(deflectionPoints.size());
	m_monoPoleTotalDelections.resize(deflectionPoints.size());
	m_pointMonoPoleDeflections.resize(deflectionPoints.size());
	m_pointDependencies.resize(deflectionPoints.size());

	for (int i = 0 ; i < deflectionPoints.size() ; i++)
	{
		// Get the monopoles which influence the deflections each of these points
		
		for (int b = 0 ; b < numBasisFunctions ; b++)
		{
			Vector2D<double> diff = deflectionPoints[i] - m_monoPoleCenters[b];

			if (diff.getLength() < m_monoPoleFunctions[b].getLengthScale())
			{
				// In this case, the basis function has an effect at the
				// point position
				
				m_pointDependencies[i].push_back(b);
			}
		}

		m_pointMonoPoleDeflections[i].resize(m_pointDependencies[i].size());

		// Calculate the deflections caused at each point by each monopole

		for (int j = 0 ; j < m_pointDependencies[i].size() ; j++)
		{
			double D_d = differenceDeflection.getLensDistance();
			PolynomialZeroMassLensParams zeroMassParams(D_d, 1.0, 
								    m_monoPoleFunctions[m_pointDependencies[i][j]].getLengthScale(),
								    m_monoPoleFunctions[m_pointDependencies[i][j]].getZeroPoint());
			PolynomialMassProfileLens lens;

			if (!lens.init(D_d, &zeroMassParams))
			{
				setErrorString(lens.getErrorString());
				return false;
			}

			Vector2D<double> diff = deflectionPoints[i] - m_monoPoleCenters[m_pointDependencies[i][j]];
			Vector2D<double> alpha;

			lens.getAlphaVector(diff, &alpha);

			m_pointMonoPoleDeflections[i][j] = alpha;
		}

		// Calculate the difference (target) deflections
		
		Vector2D<double> deflectionAngle;

		differenceDeflection.getAlphaVector(deflectionPoints[i], &deflectionAngle);

		m_differenceDeflections[i] = deflectionAngle;
	}

	return true;
}

void MonopoleGrid2::prepareFitnessCalculation()
{
	int index = 0;
	for (int y = 0 ; y < m_subDiv ; y++)
	{
		for (int x = 0 ; x < m_subDiv ; x++, index++)
		{
			const std::vector<int> &basisIndices = m_pixelDependencies[index];
			Vector2D<double> pixelCenter = m_bottomLeft + Vector2D<double>(x, y)*m_stepSize;

			double value = 0;

			for (int i = 0 ; i < basisIndices.size() ; i++)
			{
				int b = basisIndices[i];

				Vector2D<double> center = m_monoPoleCenters[b];
				Vector2D<double> diff = pixelCenter-center;

				value += m_monoPoleFunctions[b](diff) * m_monoPoleWeights[b];
			}

			m_monoPoleMap[index] = value;

			//std::cout << pixelCenter.getX() << " " << pixelCenter.getY() << " " << m_monoPoleGradientMap[index].getLength() << std::endl;
		}
		//std::cout << std::endl;
	}

	// Calculate the deflections caused by the monopoles at each point

	int numPoints = m_pointDependencies.size();

	for (int i = 0 ; i < numPoints ; i++)
	{
		const std::vector<int> &basisIndices = m_pointDependencies[i];
		Vector2D<double> totalDeflection(0, 0);
		
		for (int j = 0 ; j < basisIndices.size() ; j++)
		{
			int b = basisIndices[j];

			totalDeflection += m_pointMonoPoleDeflections[i][j] * m_monoPoleWeights[b];
		}

		m_monoPoleTotalDelections[i] = totalDeflection;
	}
}

void MonopoleGrid2::calculateFitness(double *pTotalDeflection, double *pNegative, double scale)
{
	double deflection = 0;
	double negative = 0;

	int numPixels = m_targetMap.size();

	for (int i = 0 ; i < numPixels ; i++)
	{
		double targetValue = m_targetMap[i];
		double monopoleValue = m_monoPoleMap[i] * scale;
		double totalDensity = targetValue + monopoleValue;

		if (totalDensity < 0) // try to avoid going negative
			negative += -totalDensity;
	}

	// Calculate the deflection sum -> this should be zero if we want to leave these images alone

	int numPoints = m_pointDependencies.size();

	for (int i = 0 ; i < numPoints ; i++)
	{
		Vector2D<double> sum = m_monoPoleTotalDelections[i] * scale + m_differenceDeflections[i];

		deflection += sum.getLengthSquared();
	}

	*pTotalDeflection = deflection;
	*pNegative = negative;
}

void MonopoleGrid2::clear()
{
	m_monoPoleFunctions.clear();
	m_monoPoleCenters.clear();
	m_monoPoleWeights.clear();
	m_pixelDependencies.clear();
	m_pointDependencies.clear();
	m_pointMonoPoleDeflections.clear();

	m_targetMap.clear();
	m_monoPoleMap.clear();
	m_differenceDeflections.clear();
	m_monoPoleTotalDelections.clear();

	m_subDiv = 0;
	m_stepSize = 0;
	m_bottomLeft = Vector2D<double>(0,0);
}

#define MONOPOLEGRID2_ID 0x4452474e

bool MonopoleGrid2::write(serut::SerializationInterface &si) const
{
	if (!si.writeInt32(MONOPOLEGRID2_ID))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	if (!si.writeInt32(m_subDiv))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	if (!si.writeDouble(m_stepSize))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	if (!si.writeDoubles(m_bottomLeft.getComponents(), 2))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	if (!si.writeInt32(m_monoPoleFunctions.size()))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	for (int i = 0 ; i < m_monoPoleFunctions.size() ; i++)
	{
		double values[2];

		values[0] = m_monoPoleFunctions[i].getZeroPoint();
		values[1] = m_monoPoleFunctions[i].getLengthScale();

		if (!si.writeDoubles(values, 2))
		{
			setErrorString(si.getErrorString());
			return false;
		}

		if (!si.writeDoubles(m_monoPoleCenters[i].getComponents(), 2))
		{
			setErrorString(si.getErrorString());
			return false;
		}
	}

	for (int i = 0 ; i < m_pixelDependencies.size() ; i++)
	{
		int numDepend = m_pixelDependencies[i].size();

		if (!si.writeInt32(numDepend))
		{
			setErrorString(si.getErrorString());
			return false;
		}

		for (int j = 0 ; j < numDepend ; j++)
		{
			if (!si.writeInt32(m_pixelDependencies[i][j]))
			{
				setErrorString(si.getErrorString());
				return false;
			}
		}
	}

	int32_t numPoints = m_pointDependencies.size();

	if (!si.writeInt32(numPoints)) // write the number of deflection points
	{
		setErrorString(si.getErrorString());
		return false;
	}

	for (int i = 0 ; i < numPoints ; i++)
	{
		int numDepend = m_pointDependencies[i].size();

		if (!si.writeInt32(numDepend))
		{
			setErrorString(si.getErrorString());
			return false;
		}

		for (int j = 0 ; j < numDepend ; j++)
		{
			if (!si.writeInt32(m_pointDependencies[i][j]))
			{
				setErrorString(si.getErrorString());
				return false;
			}
			if (!si.writeDoubles(m_pointMonoPoleDeflections[i][j].getComponents(), 2))
			{
				setErrorString(si.getErrorString());
				return false;
			}
		}

		if (!si.writeDoubles(m_differenceDeflections[i].getComponents(), 2))
		{
			setErrorString(si.getErrorString());
			return false;
		}
	}

	if (!si.writeDoubles(m_targetMap))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	return true;
}

bool MonopoleGrid2::read(serut::SerializationInterface &si)
{
	int32_t id;

	if (!si.readInt32(&id))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	if (id != MONOPOLEGRID2_ID)
	{
		setErrorString("Read an invalid ID");
		return false;
	}

	int32_t subDiv;

	if (!si.readInt32(&subDiv))
	{
		setErrorString("Read an invalid subdivision number");
		return false;
	}

	double stepSize;

	if (!si.readDouble(&stepSize))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	Vector2D<double> bottomLeft;

	if (!si.readDoubles(bottomLeft.getComponents(), 2))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	int32_t numBasisFunctions;

	if (!si.readInt32(&numBasisFunctions))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	std::vector<MonopoleBasisFunction> basisFunctions;
	std::vector<Vector2D<double> > centers;

	for (int32_t i = 0 ; i < numBasisFunctions ; i++)
	{
		double values[2];
		Vector2D<double> center;

		if (!si.readDoubles(values, 2))
		{
			setErrorString(si.getErrorString());
			return false;
		}

		basisFunctions.push_back(MonopoleBasisFunction(values[0], values[1]));

		if (!si.readDoubles(center.getComponents(), 2))
		{
			setErrorString(si.getErrorString());
			return false;
		}

		centers.push_back(center);
	}

	int numPixels = subDiv*subDiv;

	std::vector<std::vector<int> > dependencies(numPixels);

	for (int i = 0 ; i < numPixels ; i++)
	{
		int32_t numDepend;

		if (!si.readInt32(&numDepend))
		{
			setErrorString(si.getErrorString());
			return false;
		}

		dependencies[i].resize(numDepend);

		for (int32_t j = 0 ; j < numDepend ; j++)
		{
			int32_t value;

			if (!si.readInt32(&value))
			{
				setErrorString(si.getErrorString());
				return false;
			}

			dependencies[i][j] = value;
		}
	}

	int32_t numPoints;

	if (!si.readInt32(&numPoints))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	std::vector<std::vector<int> > pointDependencies(numPoints);
	std::vector<std::vector<Vector2D<double> > > pointMonoPoleDeflections(numPoints);
	std::vector<Vector2D<double> > targetDeflections(numPoints);

	for (int32_t i = 0 ; i < numPoints ; i++)
	{
		int32_t numDepend;

		if (!si.readInt32(&numDepend))
		{
			setErrorString(si.getErrorString());
			return false;
		}

		pointDependencies[i].resize(numDepend);
		pointMonoPoleDeflections[i].resize(numDepend);

		for (int32_t j = 0 ; j < numDepend ; j++)
		{
			int32_t value;

			if (!si.readInt32(&value))
			{
				setErrorString(si.getErrorString());
				return false;
			}

			pointDependencies[i][j] = value;

			double deflection[2];

			if (!si.readDoubles(deflection, 2))
			{
				setErrorString(si.getErrorString());
				return false;
			}

			pointMonoPoleDeflections[i][j] = Vector2D<double>(deflection[0], deflection[1]);
		}

		double targetDeflection[2];

		if (!si.readDoubles(targetDeflection, 2))
		{
			setErrorString(si.getErrorString());
			return false;
		}

		targetDeflections[i] = Vector2D<double>(targetDeflection[0], targetDeflection[1]);
	}

	std::vector<double> targetMap(numPixels);

	if (!si.readDoubles(targetMap))
	{
		setErrorString(si.getErrorString());
		return false;
	}

	// Ok, we've got it all, now store it in this instance

	clear();

	m_subDiv = subDiv;
	m_stepSize = stepSize;
	m_bottomLeft = bottomLeft;

	m_monoPoleFunctions = basisFunctions;
	m_monoPoleCenters = centers;

	m_monoPoleWeights.resize(m_monoPoleFunctions.size());
	for (int i = 0 ; i < m_monoPoleWeights.size() ; i++)
		m_monoPoleWeights[i] = 1;

	m_pixelDependencies = dependencies;
	m_pointDependencies = pointDependencies;
	m_pointMonoPoleDeflections = pointMonoPoleDeflections;
	m_targetMap = targetMap;

	m_monoPoleMap.resize(m_targetMap.size());

	m_differenceDeflections = targetDeflections;
	m_monoPoleTotalDelections.resize(m_differenceDeflections.size());

	return true;
}

} // end namespace

