/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "graleconfig.h"

#ifdef GRALECONFIG_SUPPORT_QT5

#include "imageplanepngwriter.h"
#include "imageplane.h"
#include <QImage>

#include "debugnew.h"

namespace grale
{

bool ImagePlanePNGWriter::write(ImagePlane *p, int subSamples)
{
	int numx,numy;
	
	numx = p->getNumXPixels();
	numy = p->getNumYPixels();

	QImage img(numx, numy, QImage::Format_RGB32);

	img.fill(qRgb(0,0,0));

	if (caust)
	{
		const std::list<Vector2D<double> > &caustics = p->getCausticPoints();
		std::list<Vector2D<double> >::const_iterator it;
		QRgb color = qRgb(caustR/256, caustG/256, caustB/256);

		for (it = caustics.begin() ; it != caustics.end() ; it++)
		{
			Vector2D<double> v = (*it);
			int xpos, ypos;

			if (p->getIndices(v,&xpos,&ypos))
			{
				for (int i = -linewidth/2 ; i <= linewidth/2 ; i++)
				{
					for (int j = -linewidth/2 ; j <= linewidth/2 ; j++)
					{
						int x = xpos+i;
						int y = numy-1-(ypos+j);

						if (x >= 0 && x < numx && y >= 0 && y < numy)
							img.setPixel(x,y,color);
					}
				}
			}
		}
	}
	if (crit)
	{
		const std::list<Vector2D<double> > &criticallines = p->getCriticalLinePoints();
		std::list<Vector2D<double> >::const_iterator it;
		QRgb color = qRgb(critR/256, critG/256, critB/256);

		for (it = criticallines.begin() ; it != criticallines.end() ; it++)
		{
			Vector2D<double> v = (*it);
			int xpos, ypos;

			if (p->getIndices(v,&xpos,&ypos))
			{
				for (int i = -linewidth/2 ; i <= linewidth/2 ; i++)
				{
					for (int j = -linewidth/2 ; j <= linewidth/2 ; j++)
					{
						int x = xpos+i;
						int y = numy-1-(ypos+j);

						if (x >= 0 && x < numx && y >= 0 && y < numy)
							img.setPixel(x,y,color);
					}
				}
			}
		}
	}
	if (srcs)
	{
		int x,y;

		for (y = 0 ; y < numy ; y++)
		{
			for (x = 0 ; x < numx ; x++)
			{
				double value = p->getSourceIntensityAccurate(x,y);

				if (value > 0)
				{
					if (!sourcergb)
					{
						int R = (int)(((value)/sourcedim)+0.5L);
						int G = (int)(((value)/sourcedim)+0.5L);
						int B = (int)(((value)/sourcedim)+0.5L);

						if (R > 65535)
							R = 65535;
						if (G > 65535)
							G = 65535;
						if (B > 65535)
							G = 65535;

						QRgb color = qRgb(R/256, G/256, B/256);	

						img.setPixel(x, numy-1-y, color);
					}
					else
					{
						QRgb color = qRgb(sourceR/256, sourceG/256, sourceB/256);
						img.setPixel(x,numy-1-y,color); // corner is (1,1) instead of (0,0)
					}
				}
			}
		}
	}
	if (im)
	{
		int x,y;

		for (y = 0 ; y < numy ; y++)
		{
			for (x = 0 ; x < numx ; x++)
			{
				double value = p->getImageIntensityAccurate(x,y,subSamples);

				if (value > 0)
				{
					int R = (int)(value+0.5);
					int G = (int)(value+0.5);
					int B = (int)(value+0.5);
				
					if (R > 65535)
						R = 65535;
					if (G > 65535)
						G = 65535;
					if (B > 65535)
						G = 65535;

					QRgb color = qRgb(R/256, G/256, B/256);	
					img.setPixel(x,numy-1-y,color); // corner is (1,1) instead of (0,0)
				}
			}
		}
	}
	if (!img.save(filename.c_str(), "PNG"))
	{
		setErrorString(std::string("Couldn't write PNG image to file " + filename));
		return false;
	}
	return true;
}

} // end namespace

#endif // GRALECONFIG_SUPPORT_QT5

