/*

  This file is a part of GRALE, a library to facilitate the simulation
  and inversion of gravitational lenses.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "graleconfig.h"
#include "discretesource.h"
#include "fitsimage.h"
#include "gridfunction.h"
//#include <stdio.h>
#include <string.h>

#include "debugnew.h"

namespace grale
{

DiscreteSource::DiscreteSource(Vector2D<double> angularpos, double angle, double brightnessScale) : SourceImage(SourceImage::Discrete, angularpos, angle, brightnessScale)
{
	m_pGridFunction = 0;
	m_maxRadius = 0;
}

DiscreteSource::DiscreteSource() : SourceImage(SourceImage::Discrete, Vector2D<double>(0,0), 0, 1.0)
{
	m_pGridFunction = 0;
	m_maxRadius = 0;
}

DiscreteSource::~DiscreteSource()
{
	if (m_pGridFunction)
		delete m_pGridFunction;
}

bool DiscreteSource::loadFITS(const std::string &fname, double angularWidth, double angularHeight)
{
	if (m_pGridFunction != 0)
	{
		setErrorString("Pixel values were already loaded");
		return false;
	}
	
	FITSImage fitsImage;

	if (!fitsImage.loadNoCoordinates(fname))
	{
		setErrorString(std::string("Couldn't load FITS image data: ") + fitsImage.getErrorString());
		return false;
	}

	angularWidth = ABS(angularWidth);
	angularHeight = ABS(angularHeight);

	int numPixels = fitsImage.getWidth()*fitsImage.getHeight();
	double *pPixels = new double[numPixels];

	memcpy(pPixels, fitsImage.getPixels(), numPixels*sizeof(double));
	m_pGridFunction = new GridFunction(pPixels, Vector2D<double>(-angularWidth/2.0, -angularHeight/2.0),
			                   Vector2D<double>(angularWidth/2.0, angularHeight/2.0),
					   fitsImage.getWidth(),
					   fitsImage.getHeight(), true);

	calcMaxRadius();
	return true;
}

SourceImage *DiscreteSource::createCopy() const
{
	DiscreteSource *pNewSrc = new DiscreteSource(getAngularPosition(), getAngle(), getBrightnessScale());

	int numPixels = m_pGridFunction->getWidth()*m_pGridFunction->getHeight();
	double *pPixels = new double[numPixels];
	
	memcpy(pPixels, m_pGridFunction->getValues(), numPixels*sizeof(double));
	pNewSrc->m_pGridFunction = new GridFunction(pPixels, m_pGridFunction->getBottomLeft(),
			                            m_pGridFunction->getTopRight(), m_pGridFunction->getWidth(),
						    m_pGridFunction->getHeight(), true);
	pNewSrc->calcMaxRadius();
	return pNewSrc;
}

bool DiscreteSource::setGridFunction(const GridFunction *pGridFunction)
{
	if (m_pGridFunction != 0)
	{
		setErrorString("Pixel values were already loaded");
		return false;
	}

	int numPixels = pGridFunction->getWidth()*pGridFunction->getHeight();
	double *pPixels = new double[numPixels];
	
	memcpy(pPixels, pGridFunction->getValues(), numPixels*sizeof(double));
	m_pGridFunction = new GridFunction(pPixels, pGridFunction->getBottomLeft(),
			                   pGridFunction->getTopRight(), pGridFunction->getWidth(),
					   pGridFunction->getHeight(), true);
	calcMaxRadius();
	return true;
}

double DiscreteSource::getIntensityInternal(Vector2D<double> diff) const
{
	return (*m_pGridFunction)(diff);
}

void DiscreteSource::calcMaxRadius()
{
	double x1 = m_pGridFunction->getBottomLeft().getX();
	double y1 = m_pGridFunction->getBottomLeft().getY();
	double x2 = m_pGridFunction->getTopRight().getX();
	double y2 = m_pGridFunction->getTopRight().getY();

	Vector2D<double> c1(x1,y1);
	Vector2D<double> c2(x1,y2);
	Vector2D<double> c3(x2,y1);
	Vector2D<double> c4(x2,y2);

	m_maxRadius = c1.getLength();
	m_maxRadius = MAX(m_maxRadius, c2.getLength());
	m_maxRadius = MAX(m_maxRadius, c3.getLength());
	m_maxRadius = MAX(m_maxRadius, c4.getLength());
}

} // end namespace


